# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

# TODO: add tests for all machines
# TODO: add tests for new status callbacks

"""Base classes for machine types. Do not use directly."""

import binascii
import threading

import serial

from plover import _, log
from plover.machine.keymap import Keymap
from plover.misc import boolean


# i18n: Machine state.
STATE_STOPPED = _("stopped")
# i18n: Machine state.
STATE_INITIALIZING = _("initializing")
# i18n: Machine state.
STATE_RUNNING = _("connected")
# i18n: Machine state.
STATE_ERROR = _("disconnected")


class StenotypeBase:
    """The base class for all Stenotype classes."""

    # Layout of physical keys.
    KEYS_LAYOUT = ""
    # And special actions to map to.
    ACTIONS = ()
    # Fallback to use as machine type for finding a compatible keymap
    # if one is not already available for this machine type.
    KEYMAP_MACHINE_TYPE = None

    def __init__(self):
        # Setup default keymap with no translation of keys.
        keys = self.get_keys()
        self.keymap = Keymap(keys, keys)
        self.keymap.set_mappings(zip(keys, keys))
        self.stroke_subscribers = []
        self.state_subscribers = []
        self.state = STATE_STOPPED

    def set_keymap(self, keymap):
        """Setup machine keymap."""
        self.keymap = keymap

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        pass

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        pass

    def add_stroke_callback(self, callback):
        """Subscribe to output from the stenotype machine.

        Argument:

        callback -- The function to call whenever there is output from
        the stenotype machine and output is being captured.

        """
        self.stroke_subscribers.append(callback)

    def remove_stroke_callback(self, callback):
        """Unsubscribe from output from the stenotype machine.

        Argument:

        callback -- A function that was previously subscribed.

        """
        self.stroke_subscribers.remove(callback)

    def add_state_callback(self, callback):
        self.state_subscribers.append(callback)

    def remove_state_callback(self, callback):
        self.state_subscribers.remove(callback)

    def _notify(self, steno_keys):
        """Invoke the callback of each subscriber with the given argument."""
        for callback in self.stroke_subscribers:
            callback(steno_keys)

    def _notify_keys(self, steno_keys):
        steno_keys = self.keymap.keys_to_actions(steno_keys)
        if steno_keys:
            self._notify(steno_keys)

    def set_suppression(self, enabled):
        """Enable keyboard suppression.

        This is only of use for the keyboard machine,
        to suppress the keyboard when then engine is running.
        """
        pass

    def suppress_last_stroke(self, send_backspaces):
        """Suppress the last stroke key events after the fact.

        This is only of use for the keyboard machine,
        and the engine is resumed with a command stroke.

        Argument:

        send_backspaces -- The function to use to send backspaces.
        """
        pass

    def _set_state(self, state):
        self.state = state
        for callback in self.state_subscribers:
            callback(state)

    def _stopped(self):
        self._set_state(STATE_STOPPED)

    def _initializing(self):
        self._set_state(STATE_INITIALIZING)

    def _ready(self):
        self._set_state(STATE_RUNNING)

    def _error(self):
        self._set_state(STATE_ERROR)

    @classmethod
    def get_actions(cls):
        """List of supported actions to map to."""
        return cls.ACTIONS

    @classmethod
    def get_keys(cls):
        return tuple(cls.KEYS_LAYOUT.split())

    @classmethod
    def get_option_info(cls):
        """Get the default options for this machine."""
        return {}


class ThreadedStenotypeBase(StenotypeBase, threading.Thread):
    """Base class for thread based machines.

    Subclasses should override run.
    """

    def __init__(self):
        threading.Thread.__init__(self)
        self._on_unhandled_exception(self._error)
        self.name += "-machine"
        StenotypeBase.__init__(self)
        self.finished = threading.Event()

    def _on_unhandled_exception(self, action):
        super_invoke_excepthook = self._invoke_excepthook

        def invoke_excepthook(self):
            action()
            super_invoke_excepthook(self)

        self._invoke_excepthook = invoke_excepthook

    def run(self):
        """This method should be overridden by a subclass."""
        pass

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        self.finished.clear()
        self._initializing()
        self.start()

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        self.finished.set()
        try:
            self.join()
        except RuntimeError:
            pass
        self._stopped()


class SerialStenotypeBase(ThreadedStenotypeBase):
    """For use with stenotype machines that connect via serial port.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.

    """

    # Default serial parameters.
    SERIAL_PARAMS = {
        "port": None,
        "baudrate": 9600,
        "bytesize": 8,
        "parity": "N",
        "stopbits": 1,
        "timeout": 2.0,
    }

    def __init__(self, serial_params):
        """Monitor the stenotype over a serial port.

        The key-value pairs in the <serial_params> dict are the same
        as the keyword arguments for a serial.Serial object.

        """
        ThreadedStenotypeBase.__init__(self)
        self._on_unhandled_exception(self._handle_disconnect)
        self.serial_port = None
        self.serial_params = serial_params

    def _close_port(self):
        if self.serial_port is None:
            return
        self.serial_port.close()
        self.serial_port = None

    def _handle_disconnect(self):
        self._close_port()
        self._error()

    def start_capture(self):
        self._close_port()

        try:
            self.serial_port = serial.Serial(**self.serial_params)
        except (serial.SerialException, OSError):
            log.warning("Can't open serial port", exc_info=True)
            self._error()
            return

        if not self.serial_port.isOpen():
            log.warning("Serial port is not open: %s", self.serial_params.get("port"))
            self._error()
            return

        return ThreadedStenotypeBase.start_capture(self)

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        ThreadedStenotypeBase.stop_capture(self)
        self._close_port()

    @classmethod
    def get_option_info(cls):
        """Get the default options for this machine."""
        sb = lambda s: int(float(s)) if float(s).is_integer() else float(s)
        converters = {
            "port": str,
            "baudrate": int,
            "bytesize": int,
            "parity": str,
            "stopbits": sb,
            "timeout": float,
            "xonxoff": boolean,
            "rtscts": boolean,
        }
        return {
            setting: (default, converters[setting])
            for setting, default in cls.SERIAL_PARAMS.items()
        }

    def _iter_packets(self, packet_size):
        """Yield packets of <packets_size> bytes until the machine is stopped.

        N.B.: to workaround the fact that the Toshiba Bluetooth stack
        on Windows does not correctly handle the read timeout setting
        (returning immediately if some data is already available):
        - the effective timeout is re-configured to <timeout/packet_size>
        - multiple reads are  done (until a packet is complete)
        - an incomplete packet will only be discarded if one of
          those reads return no data (but not on short read)
        """
        self.serial_port.timeout = max(
            self.serial_params.get("timeout", 1.0) / packet_size,
            0.01,
        )
        packet = b""
        while not self.finished.is_set():
            raw = self.serial_port.read(packet_size - len(packet))
            if not raw:
                if packet:
                    log.error(
                        "discarding incomplete packet: %s", binascii.hexlify(packet)
                    )
                packet = b""
                continue
            packet += raw
            if len(packet) != packet_size:
                continue
            yield packet
            packet = b""



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Configuration management."""

from collections import ChainMap, namedtuple, OrderedDict
import configparser
import json
import re

from plover.exception import InvalidConfigurationError
from plover.machine.keymap import Keymap
from plover.registry import registry
from plover.resource import resource_update
from plover.misc import boolean, expand_path, shorten_path
from plover import log


# General configuration sections, options and defaults.
MACHINE_CONFIG_SECTION = "Machine Configuration"

LEGACY_DICTIONARY_CONFIG_SECTION = "Dictionary Configuration"

LOGGING_CONFIG_SECTION = "Logging Configuration"

OUTPUT_CONFIG_SECTION = "Output Configuration"
DEFAULT_UNDO_LEVELS = 100
MINIMUM_UNDO_LEVELS = 1
DEFAULT_TIME_BETWEEN_KEY_PRESSES = 0
MINIMUM_TIME_BETWEEN_KEY_PRESSES = 0

DEFAULT_SYSTEM_NAME = "English Stenotype"

SYSTEM_CONFIG_SECTION = "System: %s"
SYSTEM_KEYMAP_OPTION = "keymap[%s]"


class DictionaryConfig(namedtuple("DictionaryConfig", "path enabled")):
    def __new__(cls, path, enabled=True):
        return super().__new__(cls, expand_path(path), enabled)

    @property
    def short_path(self):
        return shorten_path(self.path)

    def to_dict(self):
        # Note: do not use _asdict because of
        # https://bugs.python.org/issue24931
        return {
            "path": self.short_path,
            "enabled": self.enabled,
        }

    def replace(self, **kwargs):
        return self._replace(**kwargs)

    @staticmethod
    def from_dict(d):
        return DictionaryConfig(**d)

    def __repr__(self):
        return "DictionaryConfig(%r, %r)" % (self.short_path, self.enabled)


ConfigOption = namedtuple(
    "ConfigOption",
    """
                          name default
                          getter setter
                          validate full_key
                          """,
)


class InvalidConfigOption(ValueError):
    def __init__(self, raw_value, fixed_value, message=None):
        super().__init__(raw_value)
        self.raw_value = raw_value
        self.fixed_value = fixed_value
        self.message = message

    def __str__(self):
        return self.message or repr(self.raw_value)


def raw_option(name, default, section, option, validate):
    option = option or name

    def getter(config, key):
        return config._config[section][option]

    def setter(config, key, value):
        config._set(section, option, value)

    return ConfigOption(name, lambda c, k: default, getter, setter, validate, None)


def json_option(name, default, section, option, validate):
    option = option or name

    def getter(config, key):
        value = config._config[section][option]
        try:
            return json.loads(value)
        except json.JSONDecodeError as e:
            raise InvalidConfigOption(value, default) from e

    def setter(config, key, value):
        if isinstance(value, set):
            # JSON does not support sets.
            value = list(sorted(value))
        config._set(
            section, option, json.dumps(value, sort_keys=True, ensure_ascii=False)
        )

    return ConfigOption(name, default, getter, setter, validate, None)


def int_option(name, default, minimum, maximum, section, option=None):
    option = option or name

    def getter(config, key):
        return config._config[section][option]

    def setter(config, key, value):
        config._set(section, option, str(value))

    def validate(config, key, value):
        try:
            value = int(value)
        except ValueError as e:
            raise InvalidConfigOption(value, default) from e
        if (minimum is not None and value < minimum) or (
            maximum is not None and value > maximum
        ):
            message = "%s not in [%s, %s]" % (value, minimum or "-∞", maximum or "∞")
            raise InvalidConfigOption(value, default, message)
        return value

    return ConfigOption(name, lambda c, k: default, getter, setter, validate, None)


def boolean_option(name, default, section, option=None):
    option = option or name

    def getter(config, key):
        return config._config[section][option]

    def setter(config, key, value):
        config._set(section, option, str(value))

    def validate(config, key, value):
        try:
            return boolean(value)
        except ValueError as e:
            raise InvalidConfigOption(value, default) from e

    return ConfigOption(name, lambda c, k: default, getter, setter, validate, None)


def choice_option(name, choices, section, option=None):
    default = choices[0]

    def validate(config, key, value):
        if value not in choices:
            raise InvalidConfigOption(value, default)
        return value

    return raw_option(name, default, section, option, validate)


def plugin_option(name, plugin_type, default, section, option=None):
    def validate(config, key, value):
        try:
            return registry.get_plugin(plugin_type, value).name
        except KeyError as e:
            raise InvalidConfigOption(value, default) from e

    return raw_option(name, default, section, option, validate)


def opacity_option(name, section, option=None):
    return int_option(name, 100, 0, 100, section, option)


def path_option(name, default, section, option=None):
    option = option or name

    def getter(config, key):
        return expand_path(config._config[section][option])

    def setter(config, key, value):
        config._set(section, option, shorten_path(value))

    def validate(config, key, value):
        if not isinstance(value, str):
            raise InvalidConfigOption(value, default)
        return value

    return ConfigOption(name, lambda c, k: default, getter, setter, validate, None)


def enabled_extensions_option():
    def validate(config, key, value):
        if not isinstance(value, (list, set, tuple)):
            raise InvalidConfigOption(value, ())
        return set(value)

    return json_option(
        "enabled_extensions",
        lambda c, k: set(),
        "Plugins",
        "enabled_extensions",
        validate,
    )


def machine_specific_options():
    def full_key(config, key):
        if isinstance(key, tuple):
            assert len(key) == 2
            return key
        return (key, config["machine_type"])

    def default(config, key):
        machine_class = registry.get_plugin("machine", key[1]).obj
        return {
            name: params[0] for name, params in machine_class.get_option_info().items()
        }

    def getter(config, key):
        return config._config[key[1]]

    def setter(config, key, value):
        config._config[key[1]] = value

    def validate(config, key, raw_options):
        if not isinstance(raw_options, (dict, configparser.SectionProxy)):
            raise InvalidConfigOption(raw_options, default(config, key))
        machine_options = OrderedDict()
        invalid_options = OrderedDict()
        machine_class = registry.get_plugin("machine", key[1]).obj
        for name, params in sorted(machine_class.get_option_info().items()):
            fallback, convert = params
            try:
                raw_value = raw_options[name]
            except KeyError:
                value = fallback
            else:
                try:
                    value = convert(raw_value)
                except ValueError:
                    invalid_options[name] = raw_value
                    value = fallback
            machine_options[name] = value
        if invalid_options:
            raise InvalidConfigOption(invalid_options, machine_options)
        return machine_options

    return ConfigOption(
        "machine_specific_options", default, getter, setter, validate, full_key
    )


def system_keymap_option():
    def full_key(config, key):
        if isinstance(key, tuple):
            assert len(key) == 3
            return key
        return (key, config["system_name"], config["machine_type"])

    def location(config, key):
        return SYSTEM_CONFIG_SECTION % key[1], SYSTEM_KEYMAP_OPTION % key[2]

    def build_keymap(config, key, mappings=None):
        system = registry.get_plugin("system", key[1]).obj
        machine_class = registry.get_plugin("machine", key[2]).obj
        keymap = Keymap(
            machine_class.get_keys(), system.KEYS + machine_class.get_actions()
        )
        if mappings is None:
            mappings = system.KEYMAPS.get(key[2])
        if mappings is None:
            if machine_class.KEYMAP_MACHINE_TYPE is not None:
                # Try fallback.
                return build_keymap(
                    config, (key[0], key[1], machine_class.KEYMAP_MACHINE_TYPE)
                )
            # No fallback...
            mappings = {}
        keymap.set_mappings(mappings)
        return keymap

    def default(config, key):
        return build_keymap(config, key)

    def getter(config, key):
        section, option = location(config, key)
        return config._config[section][option]

    def setter(config, key, keymap):
        section, option = location(config, key)
        config._set(section, option, str(keymap))

    def validate(config, key, value):
        try:
            return build_keymap(config, key, value)
        except (TypeError, ValueError) as e:
            raise InvalidConfigOption(value, default(config, key)) from e

    return ConfigOption("system_keymap", default, getter, setter, validate, full_key)


def dictionaries_option():
    def full_key(config, key):
        if isinstance(key, tuple):
            assert len(key) == 2
            return key
        return (key, config["system_name"])

    def location(config, key):
        return (
            SYSTEM_CONFIG_SECTION % key[1],
            "dictionaries",
        )

    def default(config, key):
        system = registry.get_plugin("system", key[1]).obj
        return [DictionaryConfig(path) for path in system.DEFAULT_DICTIONARIES]

    def legacy_getter(config):
        options = config._config[LEGACY_DICTIONARY_CONFIG_SECTION].items()
        return [
            {"path": value}
            for name, value in reversed(sorted(options))
            if re.match(r"dictionary_file\d*$", name) is not None
        ]

    def getter(config, key):
        section, option = location(config, key)
        value = config._config.get(section, option, fallback=None)
        if value is None:
            return legacy_getter(config)
        return json.loads(value)

    def setter(config, key, dictionaries):
        section, option = location(config, key)
        config._set(
            section,
            option,
            json.dumps([d.to_dict() for d in dictionaries], sort_keys=True),
        )
        config._config.remove_section(LEGACY_DICTIONARY_CONFIG_SECTION)

    def validate(config, key, value):
        dictionaries = []
        for d in value:
            if isinstance(d, DictionaryConfig):
                pass
            elif isinstance(d, str):
                d = DictionaryConfig(d)
            else:
                d = DictionaryConfig.from_dict(d)
            dictionaries.append(d)
        return dictionaries

    return ConfigOption("dictionaries", default, getter, setter, validate, full_key)


class Config:
    def __init__(self, path=None):
        self._config = None
        self._cache = {}
        # A convenient place for other code to store a file name.
        self.path = path
        self.clear()

    def load(self):
        self.clear()
        with open(self.path, encoding="utf-8") as fp:
            try:
                self._config.read_file(fp)
            except configparser.Error as e:
                raise InvalidConfigurationError(str(e))

    def clear(self):
        self._config = configparser.RawConfigParser()
        self._cache.clear()

    def save(self):
        with resource_update(self.path) as temp_path:
            with open(temp_path, mode="w", encoding="utf-8") as fp:
                self._config.write(fp)

    def _set(self, section, option, value):
        if not self._config.has_section(section):
            self._config.add_section(section)
        self._config.set(section, option, value)

    # Note: order matters, e.g. machine_type comes before
    # machine_specific_options and system_keymap because
    # the latter depend on the former.
    _OPTIONS = OrderedDict(
        (opt.name, opt)
        for opt in [
            # Output.
            choice_option(
                "space_placement",
                ("Before Output", "After Output"),
                OUTPUT_CONFIG_SECTION,
            ),
            boolean_option("start_attached", False, OUTPUT_CONFIG_SECTION),
            boolean_option("start_capitalized", False, OUTPUT_CONFIG_SECTION),
            int_option(
                "undo_levels",
                DEFAULT_UNDO_LEVELS,
                MINIMUM_UNDO_LEVELS,
                None,
                OUTPUT_CONFIG_SECTION,
            ),
            int_option(
                "time_between_key_presses",
                DEFAULT_TIME_BETWEEN_KEY_PRESSES,
                MINIMUM_TIME_BETWEEN_KEY_PRESSES,
                None,
                OUTPUT_CONFIG_SECTION,
            ),
            choice_option(
                "keyboard_layout",
                ("qwerty", "qwertz", "colemak", "colemak-dh", "dvorak"),
                OUTPUT_CONFIG_SECTION,
            ),
            # Logging.
            path_option(
                "log_file_name",
                expand_path("strokes.log"),
                LOGGING_CONFIG_SECTION,
                "log_file",
            ),
            boolean_option("enable_stroke_logging", False, LOGGING_CONFIG_SECTION),
            boolean_option("enable_translation_logging", False, LOGGING_CONFIG_SECTION),
            # GUI.
            boolean_option("start_minimized", False, "Startup", "Start Minimized"),
            boolean_option("show_stroke_display", False, "Stroke Display", "show"),
            boolean_option(
                "show_suggestions_display", False, "Suggestions Display", "show"
            ),
            opacity_option("translation_frame_opacity", "Translation Frame", "opacity"),
            boolean_option("classic_dictionaries_display_order", False, "GUI"),
            # Plugins.
            enabled_extensions_option(),
            # Machine.
            boolean_option("auto_start", False, MACHINE_CONFIG_SECTION),
            plugin_option(
                "machine_type", "machine", "Keyboard", MACHINE_CONFIG_SECTION
            ),
            machine_specific_options(),
            # System.
            plugin_option(
                "system_name", "system", DEFAULT_SYSTEM_NAME, "System", "name"
            ),
            system_keymap_option(),
            dictionaries_option(),
        ]
    )

    def _lookup(self, key):
        name = key[0] if isinstance(key, tuple) else key
        opt = self._OPTIONS[name]
        if opt.full_key is not None:
            key = opt.full_key(self, key)
        return key, opt

    def __getitem__(self, key):
        key, opt = self._lookup(key)
        if key in self._cache:
            return self._cache[key]
        try:
            value = opt.validate(self, key, opt.getter(self, key))
        except (configparser.NoOptionError, KeyError):
            value = opt.default(self, key)
        except InvalidConfigOption as e:
            log.error("invalid value for %r option", opt.name, exc_info=True)
            value = e.fixed_value
        self._cache[key] = value
        return value

    def __setitem__(self, key, value):
        key, opt = self._lookup(key)
        value = opt.validate(self._config, key, value)
        opt.setter(self, key, value)
        self._cache[key] = value

    def as_dict(self):
        return {opt.name: self[opt.name] for opt in self._OPTIONS.values()}

    def update(self, **kwargs):
        new_settings = []
        new_config = ChainMap({}, self)
        for opt in self._OPTIONS.values():
            if opt.name in kwargs:
                key = opt.name
                if opt.full_key is not None:
                    key = opt.full_key(new_config, key)
                value = opt.validate(new_config, key, kwargs[opt.name])
                new_settings.append((opt, key, value))
                new_config[opt.name] = value
        for opt, key, value in new_settings:
            opt.setter(self, key, value)
            self._cache[key] = value



import locale
import os


# Note: highest priority first.
LANG_ENV_VARS = ("LC_ALL", "LC_MESSAGES", "LANG")


def get_system_language():
    try:
        return next(filter(None, map(os.environ.get, LANG_ENV_VARS)))
    except StopIteration:
        return locale.getdefaultlocale()[0]



class Output:
    """Output interface."""

    def send_backspaces(self, count):
        """Output the given number of backspaces."""
        raise NotImplementedError()

    def send_string(self, string):
        """Output the given string."""
        raise NotImplementedError()

    def send_key_combination(self, combo):
        """Output a sequence of key combinations.

        See `plover.key_combo` for the format of the `combo` string.
        """
        raise NotImplementedError()

    def set_key_press_delay(self, delay_ms):
        """Sets the delay between outputting key press events."""
        raise NotImplementedError()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for formatting.py."""

import inspect

import pytest

from plover import formatting
from plover.formatting import Case

from plover_build_utils.testing import CaptureOutput, parametrize


def action(**kwargs):
    # Support using something like `text_and_word='stuff'`
    # as a shortcut for `text='stuff', word='stuff'`.
    for k, v in list(kwargs.items()):
        if "_and_" in k:
            del kwargs[k]
            for k in k.split("_and_"):
                kwargs[k] = v
    return formatting._Action(**kwargs)


class MockTranslation:
    def __init__(self, rtfcre=tuple(), english=None, formatting=None):
        self.rtfcre = rtfcre
        self.english = english
        self.formatting = formatting

    def __str__(self):
        return str(self.__dict__)


def translation(**kwargs):
    return MockTranslation(**kwargs)


STARTING_STROKE_TESTS = (
    lambda: (
        True,
        True,
        [],
        [translation(rtfcre=("S"), english="hello")],
        None,
        ([action(prev_attach=True, text="Hello", trailing_space=" ", word="hello")],),
        [("s", "Hello")],
    ),
    lambda: (
        False,
        False,
        [],
        [translation(rtfcre=("S"), english="hello")],
        None,
        ([action(text_and_word="hello", trailing_space=" ")],),
        [("s", " hello")],
    ),
    lambda: (
        True,
        False,
        [],
        [translation(rtfcre=("S"), english="hello")],
        None,
        ([action(text="Hello", word="hello", trailing_space=" ")],),
        [("s", " Hello")],
    ),
    lambda: (
        False,
        True,
        [],
        [translation(rtfcre=("S"), english="hello")],
        None,
        ([action(text_and_word="hello", prev_attach=True, trailing_space=" ")],),
        [("s", "hello")],
    ),
)


@parametrize(STARTING_STROKE_TESTS)
def test_starting_stroke(
    capitalized, attached, undo, do, prev, expected_formats, expected_outputs
):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.start_capitalized = capitalized
    formatter.start_attached = attached
    formatter.format(undo, do, prev)
    for i, t in enumerate(do):
        assert t.formatting == expected_formats[i]
    assert output.instructions == expected_outputs


FORMATTER_TESTS = (
    lambda: (
        [translation(formatting=[action(text_and_word="hello", trailing_space=" ")])],
        [],
        None,
        (),
        [("b", 6)],
    ),
    lambda: (
        [],
        [translation(rtfcre=("S"), english="hello")],
        [
            translation(
                rtfcre=("T"),
                english="a",
                formatting=[action(text_and_word="f", trailing_space=" ")],
            )
        ],
        ([action(text_and_word="hello", trailing_space=" ")],),
        [("s", " hello")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("S"), english="hello")],
        None,
        ([action(text_and_word="hello", trailing_space=" ")],),
        [("s", " hello")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("ST-T",))],
        None,
        ([action(text_and_word="ST-T", trailing_space=" ")],),
        [("s", " ST-T")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("ST-T",))],
        [translation(formatting=[action(text_and_word="hi", trailing_space=" ")])],
        ([action(text_and_word="ST-T", trailing_space=" ")],),
        [("s", " ST-T")],
    ),
    lambda: (
        [translation(formatting=[action(text_and_word="test", trailing_space=" ")])],
        [translation(english="rest")],
        [
            translation(
                formatting=[action(next_case=Case.CAP_FIRST_WORD, trailing_space=" ")]
            )
        ],
        ([action(text="Rest", word="rest", trailing_space=" ")],),
        [("b", 4), ("s", "Rest")],
    ),
    lambda: (
        [
            translation(
                formatting=[
                    action(text_and_word="dare"),
                    action(
                        prev_attach=True, text="ing", word="daring", prev_replace="e"
                    ),
                ]
            )
        ],
        [translation(english="rest")],
        [
            translation(
                formatting=[action(next_case=Case.CAP_FIRST_WORD, trailing_space=" ")]
            )
        ],
        ([action(text="Rest", word="rest", trailing_space=" ")],),
        [("b", 6), ("s", "Rest")],
    ),
    lambda: (
        [translation(formatting=[action(text_and_word="drive", trailing_space=" ")])],
        [translation(english="driving")],
        None,
        ([action(text_and_word="driving", trailing_space=" ")],),
        [("b", 1), ("s", "ing")],
    ),
    lambda: (
        [translation(formatting=[action(text_and_word="drive", trailing_space=" ")])],
        [translation(english="{#c}driving")],
        None,
        ([action(combo="c"), action(text_and_word="driving", trailing_space=" ")],),
        [("b", 6), ("c", "c"), ("s", " driving")],
    ),
    lambda: (
        [translation(formatting=[action(text_and_word="drive", trailing_space=" ")])],
        [translation(english="{PLOVER:c}driving")],
        None,
        ([action(command="c"), action(text_and_word="driving", trailing_space=" ")],),
        [("b", 6), ("e", "c"), ("s", " driving")],
    ),
    lambda: (
        [],
        [translation(english="{PloveR:CMD}")],
        None,
        ([action(command="CMD")],),
        [("e", "CMD")],
    ),
    lambda: (
        [],
        [translation(english="{:coMManD:Cmd}")],
        None,
        ([action(command="Cmd")],),
        [("e", "Cmd")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("1",))],
        None,
        ([action(text_and_word="1", trailing_space=" ", glue=True)],),
        [("s", " 1")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("1",))],
        [translation(formatting=[action(text_and_word="hi", trailing_space=" ")])],
        ([action(text_and_word="1", trailing_space=" ", glue=True)],),
        [("s", " 1")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("1",))],
        [
            translation(
                formatting=[action(text_and_word="hi", trailing_space=" ", glue=True)]
            )
        ],
        (
            [
                action(
                    prev_attach=True,
                    text="1",
                    trailing_space=" ",
                    word="hi1",
                    glue=True,
                )
            ],
        ),
        [("s", "1")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("1-9",))],
        [
            translation(
                formatting=[action(text_and_word="hi", trailing_space=" ", glue=True)]
            )
        ],
        (
            [
                action(
                    prev_attach=True,
                    text="19",
                    trailing_space=" ",
                    word="hi19",
                    glue=True,
                )
            ],
        ),
        [("s", "19")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("ST-PL",))],
        [translation(formatting=[action(text_and_word="hi", trailing_space=" ")])],
        ([action(text_and_word="ST-PL", trailing_space=" ")],),
        [("s", " ST-PL")],
    ),
    lambda: (
        [],
        [translation(rtfcre=("ST-PL",))],
        None,
        ([action(text_and_word="ST-PL", trailing_space=" ")],),
        [("s", " ST-PL")],
    ),
)


@parametrize(FORMATTER_TESTS)
def test_formatter(undo, do, prev, expected_formats, expected_outputs):
    output = CaptureOutput()
    # Add some initial blank text so
    # undoing with no previous state
    # does not assert.
    output.text = " " * 128
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.format(undo, do, prev)
    for i, t in enumerate(do):
        assert t.formatting == expected_formats[i]
    assert output.instructions == expected_outputs


def test_action():
    assert action(word="test") != action(word="test", next_attach=True)
    assert action(text="test") == action(text="test")
    assert action(text="test", word="test").copy_state() == action(word="test")


TRANSLATION_TO_ACTIONS_TESTS = (
    lambda: ("test", action(), [action(text_and_word="test", trailing_space=" ")]),
    lambda: (
        "{^^}",
        action(),
        [
            action(
                prev_attach=True, text_and_word="", next_attach=True, orthography=False
            )
        ],
    ),
    lambda: ("1-9", action(), [action(text_and_word="1-9", trailing_space=" ")]),
    lambda: (
        "32",
        action(),
        [action(text_and_word="32", trailing_space=" ", glue=True)],
    ),
    lambda: (
        "",
        action(text_and_word="test", next_attach=True),
        [action(prev_attach=True, word="test", next_attach=True)],
    ),
    lambda: (
        "  ",
        action(text_and_word="test", next_attach=True),
        [action(prev_attach=True, word="test", next_attach=True)],
    ),
    lambda: (
        "{^} {.} hello {.} {#ALT_L(Grave)}{^ ^}",
        action(),
        [
            action(
                prev_attach=True, text_and_word="", next_attach=True, orthography=False
            ),
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(text="Hello", word="hello", trailing_space=" "),
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(
                word=".",
                trailing_space=" ",
                combo="ALT_L(Grave)",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(prev_attach=True, text=" ", word="", next_attach=True),
        ],
    ),
    lambda: (
        "{-|}{>}{&a}{>}{&b}",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(next_case=Case.LOWER_FIRST_CHAR),
            action(text_and_word="a", trailing_space=" ", glue=True),
            action(
                next_case=Case.LOWER_FIRST_CHAR, word="a", trailing_space=" ", glue=True
            ),
            action(
                prev_attach=True, text="b", word="ab", trailing_space=" ", glue=True
            ),
        ],
    ),
    lambda: (
        "{-|}{>}{&a}{>}{&b}",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(next_case=Case.LOWER_FIRST_CHAR),
            action(text_and_word="a", trailing_space=" ", glue=True),
            action(
                next_case=Case.LOWER_FIRST_CHAR, word="a", trailing_space=" ", glue=True
            ),
            action(
                prev_attach=True, text="b", word="ab", trailing_space=" ", glue=True
            ),
        ],
    ),
    lambda: (
        "{-|} equip {^s}",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(text="Equip", word="equip", trailing_space=" "),
            action(prev_attach=True, text="s", trailing_space=" ", word="equips"),
        ],
    ),
    lambda: (
        "{-|} equip {^ed}",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(text="Equip", word="equip", trailing_space=" "),
            action(prev_attach=True, text="ped", trailing_space=" ", word="equipped"),
        ],
    ),
    lambda: (
        "{>} Equip",
        action(),
        [
            action(next_case=Case.LOWER_FIRST_CHAR),
            action(text="equip", word="Equip", trailing_space=" "),
        ],
    ),
    lambda: (
        "{>} equip",
        action(),
        [
            action(next_case=Case.LOWER_FIRST_CHAR),
            action(text_and_word="equip", trailing_space=" "),
        ],
    ),
    lambda: (
        "{<} equip",
        action(),
        [
            action(next_case=Case.UPPER_FIRST_WORD),
            action(text="EQUIP", word="equip", trailing_space=" ", upper_carry=True),
        ],
    ),
    lambda: (
        "{<} EQUIP",
        action(),
        [
            action(next_case=Case.UPPER_FIRST_WORD),
            action(text_and_word="EQUIP", trailing_space=" ", upper_carry=True),
        ],
    ),
    lambda: (
        "{<} equip {^ed}",
        action(),
        [
            action(next_case=Case.UPPER_FIRST_WORD),
            action(text="EQUIP", word="equip", trailing_space=" ", upper_carry=True),
            action(
                prev_attach=True,
                text="PED",
                trailing_space=" ",
                word="equipped",
                upper_carry=True,
            ),
        ],
    ),
    lambda: (
        "equip {*-|}",
        action(),
        [
            action(text_and_word="equip", trailing_space=" "),
            action(
                prev_attach=True,
                text="Equip",
                trailing_space=" ",
                word_and_prev_replace="equip",
            ),
        ],
    ),
    lambda: (
        "equip {^ed} {*-|}",
        action(),
        [
            action(text_and_word="equip", trailing_space=" "),
            action(prev_attach=True, text="ped", trailing_space=" ", word="equipped"),
            action(
                prev_attach=True,
                text="Equipped",
                trailing_space=" ",
                word_and_prev_replace="equipped",
            ),
        ],
    ),
    lambda: (
        "Equip {*>}",
        action(),
        [
            action(text_and_word="Equip", trailing_space=" "),
            action(
                prev_attach=True,
                text="equip",
                trailing_space=" ",
                word_and_prev_replace="Equip",
            ),
        ],
    ),
    lambda: (
        "Equip {^ed} {*>}",
        action(),
        [
            action(text_and_word="Equip", trailing_space=" "),
            action(prev_attach=True, text="ped", trailing_space=" ", word="Equipped"),
            action(
                prev_attach=True,
                text="equipped",
                trailing_space=" ",
                word_and_prev_replace="Equipped",
            ),
        ],
    ),
    lambda: (
        "equip {*<}",
        action(),
        [
            action(text_and_word="equip", trailing_space=" "),
            action(
                prev_attach=True,
                text="EQUIP",
                trailing_space=" ",
                word_and_prev_replace="equip",
            ),
        ],
    ),
    lambda: (
        "equip {^ed} {*<}",
        action(),
        [
            action(text_and_word="equip", trailing_space=" "),
            action(prev_attach=True, text="ped", trailing_space=" ", word="equipped"),
            action(
                prev_attach=True,
                text="EQUIPPED",
                trailing_space=" ",
                word_and_prev_replace="equipped",
            ),
        ],
    ),
    lambda: (
        "notanumber {*($c)}",
        action(),
        [
            action(text_and_word="notanumber", trailing_space=" "),
            action(word="notanumber", trailing_space=" "),
        ],
    ),
    lambda: (
        "0 {*($c)}",
        action(),
        [
            action(text_and_word="0", trailing_space=" "),
            action(
                prev_attach=True,
                text="$0",
                word="0",
                trailing_space=" ",
                prev_replace="0",
            ),
        ],
    ),
    lambda: (
        "0.00 {*($c)}",
        action(),
        [
            action(text_and_word="0.00", trailing_space=" "),
            action(
                prev_attach=True,
                text="$0.00",
                word="0.00",
                trailing_space=" ",
                prev_replace="0.00",
            ),
        ],
    ),
    lambda: (
        "1234 {*($c)}",
        action(),
        [
            action(text_and_word="1234", trailing_space=" "),
            action(
                prev_attach=True,
                text="$1,234",
                word="1,234",
                trailing_space=" ",
                prev_replace="1234",
            ),
        ],
    ),
    lambda: (
        "1234567 {*($c)}",
        action(),
        [
            action(text_and_word="1234567", trailing_space=" "),
            action(
                prev_attach=True,
                text="$1,234,567",
                word="1,234,567",
                trailing_space=" ",
                prev_replace="1234567",
            ),
        ],
    ),
    lambda: (
        "1234.5 {*($c)}",
        action(),
        [
            action(text_and_word="1234.5", trailing_space=" "),
            action(
                prev_attach=True,
                text="$1,234.50",
                word="1,234.50",
                trailing_space=" ",
                prev_replace="1234.5",
            ),
        ],
    ),
    lambda: (
        "1234.56 {*($c)}",
        action(),
        [
            action(text_and_word="1234.56", trailing_space=" "),
            action(
                prev_attach=True,
                text="$1,234.56",
                word="1,234.56",
                trailing_space=" ",
                prev_replace="1234.56",
            ),
        ],
    ),
    lambda: (
        "1234.567 {*($c)}",
        action(),
        [
            action(text_and_word="1234.567", trailing_space=" "),
            action(
                prev_attach=True,
                text="$1,234.57",
                word="1,234.57",
                trailing_space=" ",
                prev_replace="1234.567",
            ),
        ],
    ),
    lambda: (
        "equip {^} {^ed}",
        action(),
        [
            action(text_and_word="equip", trailing_space=" "),
            action(
                prev_attach=True,
                text="",
                word="equip",
                next_attach=True,
                orthography=False,
            ),
            action(prev_attach=True, text="ed", trailing_space=" ", word="equiped"),
        ],
    ),
    lambda: (
        "{prefix^} test {^ing}",
        action(),
        [
            action(text_and_word="prefix", next_attach=True),
            action(prev_attach=True, text_and_word="test", trailing_space=" "),
            action(prev_attach=True, text="ing", trailing_space=" ", word="testing"),
        ],
    ),
    lambda: (
        "{two prefix^} test {^ing}",
        action(),
        [
            action(text="two prefix", word="prefix", next_attach=True),
            action(prev_attach=True, text_and_word="test", trailing_space=" "),
            action(prev_attach=True, text="ing", trailing_space=" ", word="testing"),
        ],
    ),
    lambda: (
        "{-|}{^|~|^}",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(prev_attach=True, text_and_word="|~|", next_attach=True),
        ],
    ),
    lambda: (
        "{-|}{~|'^}cause",
        action(),
        [
            action(next_case=Case.CAP_FIRST_WORD),
            action(text_and_word="'", next_attach=True, next_case=Case.CAP_FIRST_WORD),
            action(prev_attach=True, text="Cause", trailing_space=" ", word="cause"),
        ],
    ),
    lambda: (
        "{.}{~|'^}cuz",
        action(),
        [
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(text_and_word="'", next_attach=True, next_case=Case.CAP_FIRST_WORD),
            action(prev_attach=True, text="Cuz", trailing_space=" ", word="cuz"),
        ],
    ),
    lambda: (
        "{.}{~|'^}cause",
        action(),
        [
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(text_and_word="'", next_attach=True, next_case=Case.CAP_FIRST_WORD),
            action(prev_attach=True, text="Cause", trailing_space=" ", word="cause"),
        ],
    ),
    lambda: (
        '{.}{^~|"}heyyo',
        action(),
        [
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(
                prev_attach=True,
                text_and_word='"',
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(text="Heyyo", trailing_space=" ", word="heyyo"),
        ],
    ),
    lambda: (
        "{.}{^~|^}zshrc",
        action(),
        [
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(
                prev_attach=True,
                text_and_word="",
                next_attach=True,
                next_case=Case.CAP_FIRST_WORD,
            ),
            action(prev_attach=True, text="Zshrc", trailing_space=" ", word="zshrc"),
        ],
    ),
    lambda: (
        "{.}",
        action(),
        [
            action(
                prev_attach=True,
                text_and_word=".",
                trailing_space=" ",
                next_case=Case.CAP_FIRST_WORD,
            )
        ],
    ),
    lambda: (
        "{,}",
        action(),
        [action(prev_attach=True, text_and_word=",", trailing_space=" ")],
    ),
    lambda: (
        "test{prefix^}",
        action(),
        [
            action(text_and_word="test", trailing_space=" "),
            action(text_and_word="prefix", next_attach=True),
        ],
    ),
    lambda: (
        "{prefix^}{prefix^}",
        action(),
        [
            action(text_and_word="prefix", next_attach=True),
            action(
                prev_attach=True, text="prefix", word="prefixprefix", next_attach=True
            ),
        ],
    ),
    lambda: (
        "{prefix^}{^ing}",
        action(),
        [
            action(text_and_word="prefix", next_attach=True),
            action(prev_attach=True, text="ing", trailing_space=" ", word="prefixing"),
        ],
    ),
    lambda: (
        "{prefix^}cancel{^ing}",
        action(),
        [
            action(text_and_word="prefix", next_attach=True),
            action(prev_attach=True, text="cancel", trailing_space=" ", word="cancel"),
            action(prev_attach=True, text="ing", trailing_space=" ", word="canceling"),
        ],
    ),
)


@parametrize(TRANSLATION_TO_ACTIONS_TESTS)
def test_translation_to_actions(translation, last_action, expected):
    ctx = formatting._Context([], action())
    ctx.translated(last_action)
    assert formatting._translation_to_actions(translation, ctx) == expected


RAW_TO_ACTIONS_TESTS = (
    lambda: (
        "2-6",
        action(),
        [action(glue=True, text_and_word="26", trailing_space=" ")],
    ),
    lambda: ("2", action(), [action(glue=True, text_and_word="2", trailing_space=" ")]),
    lambda: (
        "-8",
        action(),
        [action(glue=True, text_and_word="8", trailing_space=" ")],
    ),
    lambda: (
        "-68",
        action(),
        [action(glue=True, text_and_word="68", trailing_space=" ")],
    ),
    lambda: ("S-T", action(), [action(text_and_word="S-T", trailing_space=" ")]),
)


@parametrize(RAW_TO_ACTIONS_TESTS)
def test_raw_to_actions(stroke, last_action, expected):
    ctx = formatting._Context([], action())
    ctx.translated(last_action)
    assert formatting._raw_to_actions(stroke, ctx) == expected


ATOM_TO_ACTION_TESTS = (
    lambda: (
        "{^ed}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text="ed", trailing_space=" ", word="tested"),
    ),
    lambda: (
        "{^ed}",
        action(text_and_word="carry", trailing_space=" "),
        action(
            prev_attach=True,
            text="ied",
            trailing_space=" ",
            prev_replace="y",
            word="carried",
        ),
    ),
    lambda: (
        "{^er}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text="er", trailing_space=" ", word="tester"),
    ),
    lambda: (
        "{^er}",
        action(text_and_word="carry", trailing_space=" "),
        action(
            prev_attach=True,
            text="ier",
            trailing_space=" ",
            prev_replace="y",
            word="carrier",
        ),
    ),
    lambda: (
        "{^ing}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text="ing", trailing_space=" ", word="testing"),
    ),
    lambda: (
        "{^ing}",
        action(text_and_word="begin", trailing_space=" "),
        action(prev_attach=True, text="ning", trailing_space=" ", word="beginning"),
    ),
    lambda: (
        "{^ing}",
        action(text_and_word="parade", trailing_space=" "),
        action(
            prev_attach=True,
            text="ing",
            trailing_space=" ",
            prev_replace="e",
            word="parading",
        ),
    ),
    lambda: (
        "{^s}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text="s", trailing_space=" ", word="tests"),
    ),
    lambda: (
        "{,}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text_and_word=",", trailing_space=" "),
    ),
    lambda: (
        "{:}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text_and_word=":", trailing_space=" "),
    ),
    lambda: (
        "{;}",
        action(text_and_word="test", trailing_space=" "),
        action(prev_attach=True, text_and_word=";", trailing_space=" "),
    ),
    lambda: (
        "{.}",
        action(prev_attach=True, word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text_and_word=".",
            trailing_space=" ",
            next_case=Case.CAP_FIRST_WORD,
        ),
    ),
    lambda: (
        "{?}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text_and_word="?",
            trailing_space=" ",
            next_case=Case.CAP_FIRST_WORD,
        ),
    ),
    lambda: (
        "{!}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text_and_word="!",
            trailing_space=" ",
            next_case=Case.CAP_FIRST_WORD,
        ),
    ),
    lambda: (
        "{-|}",
        action(text_and_word="test", trailing_space=" "),
        action(next_case=Case.CAP_FIRST_WORD, word="test", trailing_space=" "),
    ),
    lambda: (
        "{>}",
        action(text_and_word="test", trailing_space=" "),
        action(next_case=Case.LOWER_FIRST_CHAR, word="test", trailing_space=" "),
    ),
    lambda: (
        "{<}",
        action(text_and_word="test", trailing_space=" "),
        action(next_case=Case.UPPER_FIRST_WORD, word="test", trailing_space=" "),
    ),
    lambda: (
        "{*-|}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text="Test",
            trailing_space=" ",
            word_and_prev_replace="test",
        ),
    ),
    lambda: (
        "{*>}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True, text_and_word_and_prev_replace="test", trailing_space=" "
        ),
    ),
    lambda: (
        "{*<}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text="TEST",
            word_and_prev_replace="test",
            trailing_space=" ",
        ),
    ),
    lambda: (
        "{PLOVER:test_command}",
        action(text_and_word="test", trailing_space=" "),
        action(word="test", command="test_command", trailing_space=" "),
    ),
    lambda: (
        "{&glue_text}",
        action(text_and_word="test", trailing_space=" "),
        action(text_and_word="glue_text", trailing_space=" ", glue=True),
    ),
    lambda: (
        "{&glue_text}",
        action(text_and_word="test", trailing_space=" ", glue=True),
        action(
            prev_attach=True,
            text="glue_text",
            trailing_space=" ",
            word="testglue_text",
            glue=True,
        ),
    ),
    lambda: (
        "{&glue_text}",
        action(text_and_word="test", next_attach=True),
        action(
            prev_attach=True,
            text="glue_text",
            trailing_space=" ",
            word="glue_text",
            glue=True,
        ),
    ),
    lambda: (
        "{^attach_text}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text="attach_text",
            trailing_space=" ",
            word="testattach_text",
        ),
    ),
    lambda: (
        "{^attach_text^}",
        action(text_and_word="test", trailing_space=" "),
        action(
            prev_attach=True,
            text="attach_text",
            word="testattach_text",
            next_attach=True,
        ),
    ),
    lambda: (
        "{attach_text^}",
        action(text_and_word="test", trailing_space=" "),
        action(text_and_word="attach_text", next_attach=True),
    ),
    lambda: (
        "{#ALT_L(A)}",
        action(text_and_word="test", trailing_space=" "),
        action(combo="ALT_L(A)", trailing_space=" ", word="test"),
    ),
    lambda: (
        "text",
        action(text_and_word="test", trailing_space=" "),
        action(text_and_word="text", trailing_space=" "),
    ),
    lambda: (
        "text",
        action(text_and_word="test", trailing_space=" ", glue=True),
        action(text_and_word="text", trailing_space=" "),
    ),
    lambda: (
        "text",
        action(text_and_word="test", next_attach=True),
        action(prev_attach=True, text_and_word="text", trailing_space=" "),
    ),
    lambda: (
        "text",
        action(text_and_word="test", trailing_space=" ", next_case=Case.CAP_FIRST_WORD),
        action(text="Text", trailing_space=" ", word="text"),
    ),
    lambda: (
        "some text",
        action(text_and_word="test", trailing_space=" "),
        action(text="some text", trailing_space=" ", word="text"),
    ),
    lambda: (
        "some text",
        action(
            text_and_word="test", trailing_space=" ", case=Case.TITLE, space_char=""
        ),
        action(text="SomeText", word="text", case=Case.TITLE, space_char=""),
    ),
    lambda: (
        "some text",
        action(
            text_and_word="test",
            trailing_space=" ",  # This is camel case
            case=Case.TITLE,
            space_char="",
            next_case=Case.LOWER_FIRST_CHAR,
        ),
        action(text="someText", word="text", case=Case.TITLE, space_char=""),
    ),
    lambda: (
        "some text",
        action(text_and_word="test", trailing_space=" ", space_char="_"),
        action(text="some_text", trailing_space="_", word="text", space_char="_"),
    ),
    lambda: (
        "some text",
        action(text_and_word="test", trailing_space=" ", case=Case.UPPER),
        action(text="SOME TEXT", trailing_space=" ", word="text", case=Case.UPPER),
    ),
    lambda: (
        "sOme TexT",
        action(text_and_word="test", trailing_space=" ", case=Case.LOWER),
        action(text="some text", trailing_space=" ", word="TexT", case=Case.LOWER),
    ),
    lambda: (
        "sOme TexT",
        action(text_and_word="test", trailing_space=" ", case=Case.TITLE),
        action(text="Some Text", trailing_space=" ", word="TexT", case=Case.TITLE),
    ),
    lambda: (
        "{MODE:CAPS}",
        action(text_and_word="test", trailing_space=" "),
        action(word="test", trailing_space=" ", case=Case.UPPER),
    ),
    lambda: (
        "{MODE:LOWER}",
        action(text_and_word="test", trailing_space=" "),
        action(word="test", trailing_space=" ", case=Case.LOWER),
    ),
)


@parametrize(ATOM_TO_ACTION_TESTS)
def test_atom_to_action(atom, last_action, expected):
    ctx = formatting._Context((), last_action)
    ctx.translated(last_action)
    assert formatting._atom_to_action(atom, ctx) == expected


CHANGE_MODE_TESTS = (
    # Invalid modes.
    lambda: ("", action(), ValueError),
    lambda: ("ABCD", action(), ValueError),
    # CAPS: Uppercase
    lambda: ("CAPS", action(), action(case=Case.UPPER)),
    # LOWER: Lowercase
    lambda: ("LOWER", action(), action(case=Case.LOWER)),
    # TITLE: Titlecase
    lambda: ("TITLE", action(), action(case=Case.TITLE)),
    # CAMEL: Titlecase without space
    lambda: (
        "CAMEL",
        action(),
        action(case=Case.TITLE, space_char="", next_case=Case.LOWER_FIRST_CHAR),
    ),
    # SNAKE: Underscore space
    lambda: ("SNAKE", action(), action(space_char="_")),
    # RESET_SPACE: Default space
    lambda: ("RESET_SPACE", action(space_char="ABCD"), action()),
    # RESET_CASE: No case
    lambda: ("RESET_CASE", action(case=Case.UPPER), action()),
    # SET_SPACE:xy: Set space to xy
    lambda: ("SET_SPACE:", action(space_char="test"), action(space_char="")),
    lambda: ("SET_SPACE:-", action(space_char="test"), action(space_char="-")),
    lambda: (
        "SET_SPACE:123 45",
        action(space_char="test"),
        action(space_char="123 45"),
    ),
    # RESET: No case, default space
)


@parametrize(CHANGE_MODE_TESTS)
def test_meta_mode(meta, last_action, expected):
    atom = "{MODE:" + meta + "}"
    ctx = formatting._Context((), action())
    ctx.translated(last_action)
    if inspect.isclass(expected):
        with pytest.raises(expected):
            formatting._atom_to_action(atom, ctx)
    else:
        assert formatting._atom_to_action(atom, ctx) == expected


last_action_normal = action()
last_action_capitalized = action(next_case=Case.CAP_FIRST_WORD)
last_action_attached = action(next_attach=True)

META_CARRY_CAPITALIZE_TESTS = (
    # Test word handling and space handling, standard.
    lambda: (
        "~|*",
        last_action_normal,
        (action(word="*", text="*", trailing_space=" ")),
    ),
    # With attach flags:
    lambda: (
        "~|*^",
        last_action_normal,
        (action(word="*", text="*", next_attach=True)),
    ),
    lambda: (
        "^~|*",
        last_action_normal,
        (action(word="*", text="*", trailing_space=" ", prev_attach=True)),
    ),
    lambda: (
        "^~|*^",
        last_action_normal,
        (action(word="*", text="*", prev_attach=True, next_attach=True)),
    ),
    # Should 'do nothing'.
    lambda: ("~|", last_action_capitalized, (last_action_capitalized)),
    # Should lose 'next_attach' flag.
    lambda: ("~|", last_action_attached, (action(prev_attach=True))),
    # Verify capitalize carry.
    lambda: (
        "^~|^",
        last_action_capitalized,
        (
            action(
                next_case=Case.CAP_FIRST_WORD,
                text_and_word="",
                prev_attach=True,
                next_attach=True,
            )
        ),
    ),
    lambda: (
        "^~|aset^",
        last_action_capitalized,
        (
            action(
                next_case=Case.CAP_FIRST_WORD,
                prev_attach=True,
                next_attach=True,
                text="Aset",
                word="aset",
            )
        ),
    ),
    lambda: (
        "~|aset",
        last_action_capitalized,
        (
            action(
                next_case=Case.CAP_FIRST_WORD,
                text="Aset",
                trailing_space=" ",
                word="aset",
            )
        ),
    ),
    # Verify 'next_attach' flag overriding.
    lambda: (
        "~|aset",
        last_action_attached,
        (action(prev_attach=True, text_and_word="aset", trailing_space=" ")),
    ),
    lambda: (
        "~|aset^",
        last_action_attached,
        (action(prev_attach=True, text_and_word="aset", next_attach=True)),
    ),
)


@parametrize(META_CARRY_CAPITALIZE_TESTS)
def test_meta_carry_capitalize(meta, last_action, expected):
    ctx = formatting._Context((), action())
    ctx.translated(last_action)
    assert formatting._atom_to_action("{" + meta + "}", ctx) == expected


def _apply_case_tests():
    test = " some test "
    test2 = "test Me"
    test3 = " SOME TEST "
    return (
        # INVALID
        lambda: (test, "", False, ValueError),
        lambda: (test, "TEST", False, ValueError),
        # NO-OP
        lambda: (test, None, False, test),
        lambda: (test, None, True, test),
        # TITLE
        lambda: (test, Case.TITLE, False, " Some Test "),
        # TITLE will not affect appended output
        lambda: (test, Case.TITLE, True, " some test "),
        lambda: (test2, Case.TITLE, True, "test Me"),
        # LOWER
        lambda: (test, Case.LOWER, False, " some test "),
        lambda: (test3, Case.LOWER, False, " some test "),
        lambda: (test2, Case.LOWER, True, "test me"),
        # UPPER
        lambda: (test.upper(), Case.UPPER, False, " SOME TEST "),
        lambda: (test3, Case.UPPER, False, " SOME TEST "),
        lambda: (test2, Case.UPPER, True, "TEST ME"),
    )


@parametrize(_apply_case_tests())
def test_apply_case(input_text, case, appended, expected):
    if inspect.isclass(expected):
        with pytest.raises(expected):
            formatting.apply_mode_case(input_text, case, appended)
    else:
        assert formatting.apply_mode_case(input_text, case, appended) == expected


def _apply_space_char_tests():
    test = " some text "
    test2 = "don't"
    return (
        lambda: (test, "_", "_some_text_"),
        lambda: (test, "", "sometext"),
        lambda: (test2, "_", test2),
        lambda: (test2, "", test2),
    )


@parametrize(_apply_space_char_tests())
def test_apply_space_char(text, space_char, expected):
    assert formatting.apply_mode_space_char(text, space_char) == expected


@parametrize(
    (
        lambda: ("", None),
        lambda: ("{abc}", "abc"),
        lambda: ("abc", None),
    )
)
def test_get_meta(atom, meta):
    assert formatting._get_meta(atom) == meta


@parametrize(
    (
        lambda: ("abc", "{&abc}"),
        lambda: ("1", "{&1}"),
    )
)
def test_glue_translation(s, expected):
    assert formatting._glue_translation(s) == expected


@parametrize(
    (
        lambda: ("", ""),
        lambda: ("abc", "abc"),
        lambda: (r"\{", "{"),
        lambda: (r"\}", "}"),
        lambda: (r"\{abc\}}{", "{abc}}{"),
    )
)
def test_unescape_atom(atom, text):
    assert formatting._unescape_atom(atom) == text


@parametrize(
    (
        lambda: ("", ""),
        lambda: ("abc", "Abc"),
        lambda: ("ABC", "ABC"),
    )
)
def test_capitalize_first_word(s, expected):
    assert formatting.capitalize_first_word(s) == expected


RIGHTMOST_WORD_TESTS = (
    lambda: ("", ""),
    lambda: ("\n", ""),
    lambda: ("\t", ""),
    lambda: ("abc", "abc"),
    lambda: ("a word", "word"),
    lambda: ("word.", "."),
    lambda: ("word ", ""),
    lambda: ("word\n", ""),
    lambda: ("word\t", ""),
    lambda: (" word", "word"),
    lambda: ("\nword", "word"),
    lambda: ("\tword", "word"),
)


@parametrize(RIGHTMOST_WORD_TESTS)
def test_rightmost_word(s, expected):
    assert formatting.rightmost_word(s) == expected


REPLACE_TESTS = (
    # Check that 'prev_replace' does not unconditionally erase
    # the previous character if it does not match.
    lambda: (
        [
            translation(english="{MODE:SET_SPACE:}"),
            translation(english="foobar"),
            translation(english="{^}{#Return}{^}{-|}"),
        ],
        [("s", "foobar"), ("c", "Return")],
    ),
    # Check 'prev_replace' correctly takes into account
    # the previous translation.
    lambda: (
        [
            translation(english="test "),
            translation(english="{^,}"),
        ],
        [("s", "test "), ("b", 1), ("s", ", ")],
    ),
    # While the previous translation must be taken into account,
    # any meta-command must not be fired again.
    lambda: (
        [
            translation(english="{#Return}"),
            translation(english="test"),
        ],
        [("c", "Return"), ("s", "test ")],
    ),
)


@parametrize(REPLACE_TESTS)
def test_replace(translations, expected_instructions):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.set_space_placement("After Output")
    prev = []
    for t in translations:
        formatter.format([], [t], prev)
        prev.append(t)
    assert output.instructions == expected_instructions


def test_undo_replace():
    # Undoing a replace....
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.set_space_placement("After Output")
    prev = [translation(english="test")]
    formatter.format([], prev, None)
    undo = translation(english="{^,}")
    formatter.format([], [undo], prev)
    # Undo.
    formatter.format([undo], [], prev)
    assert output.instructions == [
        ("s", "test "),
        ("b", 1),
        ("s", ", "),
        ("b", 2),
        ("s", " "),
    ]


OUTPUT_OPTIMISATION_TESTS = (
    # No change.
    lambda: (
        [
            translation(english="noop"),
        ],
        [
            translation(english="noop"),
        ],
        [("s", " noop")],
    ),
    # Append only.
    lambda: (
        [
            translation(english="test"),
        ],
        [
            translation(english="testing"),
        ],
        [("s", " test"), ("s", "ing")],
    ),
    # Chained meta-commands.
    lambda: (
        [
            translation(english="{#a}"),
        ],
        [
            translation(english="{#a}{#b}"),
        ],
        [("c", "a"), ("c", "b")],
    ),
)


@parametrize(OUTPUT_OPTIMISATION_TESTS)
def test_output_optimization(undo, do, expected_instructions):
    output = CaptureOutput()
    formatter = formatting.Formatter()
    formatter.set_output(output)
    formatter.format([], undo, None)
    formatter.format(undo, do, None)
    assert output.instructions == expected_instructions


class TestRetroFormatter:
    def setup_method(self):
        self.formatter = formatting.Formatter()
        self.translations = []
        self.retro_formatter = formatting.RetroFormatter(self.translations)

    def format(self, text):
        t = translation(english=text)
        self.formatter.format([], [t], self.translations)
        self.translations.append(t)
        return t

    ITER_LAST_ACTIONS_TESTS = (
        lambda: (
            ["Luca", "mela"],
            [
                action(text_and_word="mela", trailing_space=" "),
                action(text_and_word="Luca", trailing_space=" "),
            ],
        ),
        lambda: (
            ["{Luca^}", "{^mela}"],
            [
                action(
                    prev_attach=True, text="mela", trailing_space=" ", word="Lucamela"
                ),
                action(text_and_word="Luca", next_attach=True),
            ],
        ),
        lambda: (
            ["Luca", "{^ ^}", "mela"],
            [
                action(text_and_word="mela", trailing_space=" ", prev_attach=True),
                action(text=" ", word="", prev_attach=True, next_attach=True),
                action(text_and_word="Luca", trailing_space=" "),
            ],
        ),
        lambda: (
            ["Luca", "{-|}", "mela"],
            [
                action(text="Mela", trailing_space=" ", word="mela"),
                action(word="Luca", trailing_space=" ", next_case=Case.CAP_FIRST_WORD),
                action(text_and_word="Luca", trailing_space=" "),
            ],
        ),
    )

    @parametrize(ITER_LAST_ACTIONS_TESTS)
    def test_iter_last_actions(self, translation_list, action_list):
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_actions()) == action_list

    ITER_LAST_FRAGMENTS_TESTS = (
        lambda: (False, ["Luca", "mela"], ["mela", "Luca "]),
        lambda: (False, ["Luca{^ ^}mela"], ["mela", "Luca "]),
        lambda: (False, ["Luca, mela."], ["mela.", "Luca, "]),
        lambda: (False, ["Luca{-|}mela"], ["Mela", "Luca "]),
        lambda: (True, ["Luca{-|}mela"], ["Mela", "Luca "]),
    )

    @parametrize(ITER_LAST_FRAGMENTS_TESTS)
    def test_iter_last_fragments(self, spaces_after, translation_list, fragment_list):
        if spaces_after:
            self.formatter.set_space_placement("After Output")
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_fragments()) == fragment_list

    ITER_LAST_WORDS_TESTS = (
        lambda: (False, ["Luca", "mela"], ["mela", "Luca "]),
        lambda: (False, ["Luca{^ ^}mela"], ["mela", "Luca "]),
        lambda: (False, ["Luca, mela."], [".", "mela", ", ", "Luca"]),
        lambda: (False, ["Luca{-|}mela"], ["Mela", "Luca "]),
        lambda: (True, ["Luca{-|}mela"], ["Mela", "Luca "]),
    )

    @parametrize(ITER_LAST_WORDS_TESTS)
    def test_iter_last_words(self, spaces_after, translation_list, word_list):
        if spaces_after:
            self.formatter.set_space_placement("After Output")
        for t in translation_list:
            self.format(t)
        assert list(self.retro_formatter.iter_last_words()) == word_list

    LAST_TEXT_TESTS = (
        lambda: (False, ["Luca{-|}mela"], 3, "ela"),
        lambda: (False, ["Luca{-|}mela"], 5, " Mela"),
        lambda: (False, ["Luca{-|}mela"], 12, "Luca Mela"),
        lambda: (False, ["Luca{-|}mela"], 20, "Luca Mela"),
        lambda: (True, ["Luca{-|}mela"], 6, "a Mela"),
    )

    @parametrize(LAST_TEXT_TESTS)
    def test_last_text(self, spaces_after, translation_list, count, text):
        if spaces_after:
            self.formatter.set_space_placement("After Output")
        for t in translation_list:
            self.format(t)
        assert self.retro_formatter.last_text(count) == text



from PySide6.QtWidgets import QDialogButtonBox, QDialog

from plover.gui_qt.console_widget import ConsoleWidget
from plover.gui_qt.run_dialog_ui import Ui_RunDialog


class RunDialog(QDialog, Ui_RunDialog):
    def __init__(self, run_args, popen=None):
        super().__init__()
        self.setupUi(self)
        self._console = ConsoleWidget(popen)
        self.layout().replaceWidget(self.console, self._console)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Close).setHidden(True)
        self._console.processFinished.connect(self.on_process_finished)
        self._console.run(run_args)
        self._successful = None

    def on_process_finished(self, returncode):
        self._successful = returncode == 0
        self.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setHidden(True)
        self.buttonBox.button(QDialogButtonBox.StandardButton.Close).setHidden(False)

    def reject(self):
        if self._successful is not None:
            super().done(
                getattr(
                    QDialog.DialogCode, "Accepted" if self._successful else "Rejected"
                )
            )
            return
        self._console.terminate()


if __name__ == "__main__":
    import sys
    from PySide6.QtWidgets import QApplication

    app = QApplication([])
    dlg = RunDialog(sys.argv[1:])
    dlg.show()
    app.exec()



# -*- coding: utf-8 -*-

import pytest

from plover import system
from plover.registry import Registry
from plover.system import english_stenotype

from plover_build_utils.testing import blackbox_test


@pytest.fixture
def with_melani_system(monkeypatch, request):
    class Melani:
        # fmt: off
        KEYS = (
            "#",
            "S-", "P-", "C-", "T-", "H-", "V-", "R-",
            "I-", "A-",
            "-E", "-O",
            "-c", "-s", "-t", "-h", "-p", "-r",
            "*",
            "-i", "-e", "-a", "-o",
        )
        # fmt: on
        IMPLICIT_HYPHEN_KEYS = KEYS
        SUFFIX_KEYS = ()
        NUMBER_KEY = "#"
        NUMBERS = {
            "S-": "1-",
            "P-": "2-",
            "T-": "3-",
            "V-": "4-",
            "I-": "5-",
            "-O": "0-",
            "-c": "-6",
            "-t": "-7",
            "-p": "-8",
            "-i": "-9",
        }
        UNDO_STROKE_STENO = "*"
        ORTHOGRAPHY_RULES = []
        ORTHOGRAPHY_RULES_ALIASES = {}
        ORTHOGRAPHY_WORDLIST = None
        KEYMAPS = {}
        DICTIONARIES_ROOT = None
        DEFAULT_DICTIONARIES = ()

    registry = Registry()
    registry.register_plugin("system", "English Stenotype", english_stenotype)
    registry.register_plugin("system", "Melani", Melani)
    old_system_name = system.NAME
    with monkeypatch.context() as mp:
        mp.setattr("plover.system.registry", registry)
        yield
    system.setup(old_system_name)


@pytest.fixture
def with_korean_system(monkeypatch, request):
    class KoreanCAS:
        # fmt: off
        KEYS = (
            "1-", "2-", "3-", "4-", "5-",
            "ㅎ-", "ㅁ-", "ㄱ-", "ㅈ-", "ㄴ-",
            "ㄷ-", "ㅇ-", "ㅅ-", "ㅂ-", "ㄹ-",
            "ㅗ-", "ㅏ-", "ㅜ-",
            "-*", "-ㅓ", "-ㅣ",
            "-6", "-7", "-8", "-9", "-0",
            "-ㅎ", "-ㅇ", "-ㄹ", "-ㄱ", "-ㄷ",
            "-ㅂ", "-ㄴ", "-ㅅ", "-ㅈ", "-ㅁ",
        )
        # fmt: on
        IMPLICIT_HYPHEN_KEYS = (
            "ㅗ-",
            "ㅏ-",
            "ㅜ-",
            "-*",
            "-ㅓ",
            "-ㅣ",
        )
        SUFFIX_KEYS = ()
        NUMBER_KEY = None
        NUMBERS = {}
        UNDO_STROKE_STENO = "-ㅂㄴ"
        ORTHOGRAPHY_RULES = []
        ORTHOGRAPHY_RULES_ALIASES = {}
        ORTHOGRAPHY_WORDLIST = None
        KEYMAPS = {}
        DICTIONARIES_ROOT = None
        DEFAULT_DICTIONARIES = ()

    registry = Registry()
    registry.register_plugin("system", "English Stenotype", english_stenotype)
    registry.register_plugin("system", "Korean Modern C", KoreanCAS)
    old_system_name = system.NAME
    with monkeypatch.context() as mp:
        mp.setattr("plover.system.registry", registry)
        yield
    system.setup(old_system_name)


@blackbox_test
class TestsBlackbox:
    def test_translator_state_handling(self):
        # Check if the translator curtailing the list of last translations
        # according to its dictionary longest key does no affect things
        # like the restrospective repeat-last-stroke command.
        r"""
        "TEFT": "test",
        "R*S": "{*+}",

        TEFT/R*S  " test test"
        """

    def test_force_lowercase_title(self):
        r"""
        "T-LT": "{MODE:title}",
        "TEFT": "{>}test",

        T-LT/TEFT  " test"
        """

    def test_bug471(self):
        # Repeat-last-stroke after typing two numbers outputs the numbers
        # reversed for some combos.
        r"""
        "R*S": "{*+}",

        12/R*S  " 1212"
        """

    def test_bug535(self):
        # Currency formatting a number with a decimal fails by not erasing
        # the previous output.
        r"""
        "P-P": "{^.^}",
        "KR*UR": "{*($c)}",

        1/P-P/2/KR*UR  " $1.20"
        """

    def test_bug606(self):
        r"""
        "KWEGS": "question",
        "-S": "{^s}",
        "TP-PL": "{.}",

        :spaces_after
        KWEGS/-S/TP-PL  "questions. "
        """

    def test_bug535_spaces_after(self):
        # Currency formatting a number with a decimal fails by not erasing
        # the previous output.
        r"""
        "P-P": "{^.^}",
        "KR*UR": "{*($c)}",

        :spaces_after
        1/P-P/2/KR*UR  "$1.20 "
        """

    def test_undo_not_defined(self):
        r"""
        'TEFT': 'test',

        TEFT  ' test'
        *     ''
        """

    def test_undo_overriden(self):
        r"""
        'TEFT': 'test',
        '*': 'not undo',

        TEFT  ' test'
        *     ' test not undo'
        """

    def test_undo_macro(self):
        r"""
        'TEFT': 'test',
        'TPH-D': '=undo',

        TEFT   ' test'
        TPH-D  ''
        """

    def test_undo_fingerspelling_1(self):
        r"""
        'T*': '{>}{&t}',
        '*E': '{>}{&e}',
        'S*': '{>}{&s}',
        'T*': '{>}{&t}',

        T*/*E/S*/T*  ' test'
        */*          ' te'
        */*          ''
        """

    def test_undo_fingerspelling_2(self):
        r"""
        'T*': '{>}{&t}',
        '*E': '{>}{&e}',
        'S*': '{>}{&s}',
        'T*': '{>}{&t}',

        :spaces_after
        T*/*E/S*/T*  'test '
        */*          'te '
        */*          ''
        """

    def test_undo_replaced_1(self):
        r"""
        "TEFT": "test",
        "TEFT/TKPWOEFT": "{}",

        TEFT      ' test'
        TKPWOEFT  ''
        *         ' test'
        """

    def test_undo_replaced_2(self):
        r"""
        "HROS": "loss",
        "HRO*S": "lost",
        "*P": "{*}",

        HROS   ' loss'
        *P     ' lost'
        *      ''
        """

    def test_undo_replaced_3(self):
        r"""
        'PER': 'perfect',
        'SWAEUGS': 'situation',
        'PER/SWAEUGS': 'persuasion',
        'SP*': '{*?}',

        PER      ' perfect'
        SWAEUGS  ' persuasion'
        SP*      ' perfect situation'
        *        ' persuasion'
        """

    def test_bug557(self):
        # Using the asterisk key to delete letters in fingerspelled words
        # occasionally causes problems when the space placement is set to
        # "After Output".
        r"""
        "EU": "I",
        "HRAOEUBG": "like",
        "T*": "{>}{&t}",
        "A*": "{>}{&a}",
        "KR*": "{>}{&c}",
        "O*": "{>}{&o}",
        "S*": "{>}{&s}",

        :spaces_after
        EU/HRAOEUBG/T*/A*/KR*/O*/S*/*/*/*  "I like ta "
        """

    def test_bug557_resumed(self):
        # Using the asterisk key to delete letters in fingerspelled words
        # occasionally causes problems when the space placement is set to
        # "After Output".
        r"""
        "EU": "I",
        "HRAOEUBG": "like",
        "T*": "{>}{&t}",
        "A*": "{>}{&a}",
        "KR*": "{>}{&c}",
        "O*": "{>}{&o}",
        "S*": "{>}{&s}",

        :spaces_after
        EU/HRAOEUBG/T*/A*/KR*/O*/S*/*/*/*/*/*/HRAOEUBG  "I like like "
        """

    def test_bug557_capitalized(self):
        # Using the asterisk key to delete letters in fingerspelled words
        # occasionally causes problems when the space placement is set to
        # "After Output".
        r"""
        "EU": "I",
        "HRAOEUBG": "like",
        "T*": "{-|}{&t}",
        "A*": "{-|}{&a}",
        "KR*": "{-|}{&c}",
        "O*": "{-|}{&o}",
        "S*": "{-|}{&s}",

        :spaces_after
        EU/HRAOEUBG/T*/A*/KR*/O*/S*/*/*/*  "I like TA "
        """

    def test_capitalized_fingerspelling_spaces_after(self):
        r"""
        "HRAOEUBG": "like",
        "T*": "{&T}",
        "A*": "{&A}",
        "KR*": "{&C}",
        "O*": "{&O}",
        "S*": "{&S}",

        :spaces_after
        HRAOEUBG/T*/A*/KR*/O*/S*  "like TACOS "
        """

    def test_special_characters(self):
        r"""
        "R-R": "{^}\n{^}",
        "TAB": "\t",

        R-R/TAB  "\n\t"
        """

    def test_automatic_suffix_keys_1(self):
        r"""
        "RAEUS": "race",
        "RAEUZ": "raise",
        "-S": "{^s}",
        "-Z": "{^s}",

        RAEUSZ  " races"
        """

    def test_bug719(self):
        # Glue (&) does not work with "Spaces After".
        r"""
        "P*": "{&P}"

        :spaces_after
        P*/P*/P*/P*/P*/P*  'PPPPPP '
        """

    def test_bug741(self):
        # Uppercase last word also uppercases next word's prefix.
        r"""
        "KPA*TS": "{*<}",
        "TPAO": "foo",
        "KAUPB": "{con^}",

        TPAO/KPA*TS/KAUPB/TPAO  " FOO confoo"
        """

    def test_carry_capitalization_spacing1(self):
        r"""
        'S-P': '{^ ^}',
        'R-R': '{^~|\n^}',

        S-P/R-R  ' \n'
        """

    def test_carry_capitalization_spacing2(self):
        r"""
        'S-P': '{^ ^}',
        'R-R': '{^~|\n^}',

        :spaces_after
        S-P/R-R  ' \n'
        """

    def test_orthography1(self):
        r"""
        'TEFT': 'test',
        '-G': '{^ing}',

        TEFT/-G  ' testing'
        """

    def test_orthography2(self):
        r"""
        'SKEL': 'cancel',
        '-G': '{^ing}',

        SKEL/-G  " canceling"
        """

    def test_orthography3(self):
        r"""
        'PREPB': '{(^}',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PREPB/SKEL/-G  " (canceling"
        """

    def test_orthography4(self):
        r"""
        'SKEL': '{&c}{&a}{&n}{&c}{&e}{&l}',
        '-G': '{^ing}',

        SKEL/-G  " canceling"
        """

    def test_orthography5(self):
        r"""
        'TPAOEUPL': 'time',
        '-G': '{^ing}',

        TPAOEUPL/-G  ' timing'
        """

    def test_orthography6(self):
        r"""
        'PRE': '{prefix-^}',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PRE/SKEL/-G  " prefix-canceling"
        """

    def test_orthography7(self):
        r"""
        'PRE': '{prefix^}',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PRE/SKEL/-G  " prefixcanceling"
        """

    def test_orthography8(self):
        r"""
        'PRE': '{&p}{&r}{&e}{&f}{&i}{&x}',
        "TK-LS": "{^}",
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PRE/TK-LS/SKEL/-G  " prefixcanceling"
        """

    def test_orthography9(self):
        r"""
        'PRE': '{prefix^}',
        'SKEL': '{&c}{&a}{&n}{&c}{&e}{&l}',
        '-G': '{^ing}',

        PRE/SKEL/-G  " prefixcanceling"
        """

    def test_orthography10(self):
        r"""
        'PHO*D': "{MODE:CAMEL}",
        'PRE': '{prefix^}',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/PRE/SKEL/-G  "prefixcanceling"
        """

    def test_orthography11(self):
        r"""
        'PHO*D': "{MODE:SNAKE}",
        'TEFT': 'test',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/TEFT  "_test"
        SKEL/-G     "_test_canceling"
        """

    def test_orthography12(self):
        r"""
        'PHO*D': "{MODE:CAMEL}",
        'TEFT': 'test',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/TEFT  "test"
        SKEL/-G     "testCanceling"
        """

    def test_orthography13(self):
        r"""
        'PHO*D': "{MODE:SNAKE}",
        'PRE': '{prefix^}',
        'TEFT': 'test',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/TEFT  "_test"
        PRE/SKEL/-G     "_test_prefixcanceling"
        """

    def test_orthography14(self):
        r"""
        'PHO*D': "{MODE:CAMEL}",
        'PRE': '{prefix^}',
        'TEFT': 'test',
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/TEFT  "test"
        PRE/SKEL/-G     "testPrefixcanceling"
        """

    def test_orthography15(self):
        r"""
        'PHO*D': "{MODE:CAPS}",
        'SKEL': 'cancel',
        '-G': '{^ing}',

        PHO*D/SKEL/-G  " CANCELING"
        """

    def test_orthography16(self):
        r"""
        'SKEL': '{&C}{&A}{&N}{&C}{&E}{&L}',
        '-G': '{^ing}',

        SKEL/-G  " CANCELLing"
        """

    def test_orthography17(self):
        r"""
        'SKEL': 'CANCEL',
        '-G': '{^ing}',

        SKEL/-G  " CANCELLing"
        """

    def test_orthography18(self):
        r"""
        'TPAOEUPL': 'TIME',
        '-G': '{^ing}',

        TPAOEUPL/-G  ' TIMing'
        """

    def test_orthography19(self):
        r"""
        'TPAOEUPL': '{&T}{&I}{&M}{&E}',
        '-G': '{^ing}',

        TPAOEUPL/-G  ' TIMing'
        """

    def test_after_initial_being(self):
        r"""
        '-B': 'be',
        '-B/-G': 'being',

        :spaces_after
        -B/-G  'being '
        """

    def test_after_period(self):
        r"""
        'TEFT': 'test',
        'P-P': '{.}',

        :spaces_after
        TEFT/P-P/TEFT  'test. Test '
        """

    def test_spaces_after1(self):
        r"""
        'TEFT': 'test',
        'TEFTD': 'tested',
        '-G': '{^ing}',
        'P-P': '{.}',

        :spaces_after
        TEFT   'test '
        -G     'testing '
        P-P    'testing. '
        TEFTD  'testing. Tested '
        """

    def test_start_attached1(self):
        r"""
        'TEFT': 'test',

        :start_attached
        TEFT  'test'
        """

    def test_start_attached2(self):
        r"""
        'TEFT': 'test',

        :start_attached
        :spaces_after
        TEFT  'test '
        """

    def test_space_placement_change1(self):
        r"""
        'TEFT': 'test',
        'TEFTD': 'tested',
        'TEFTS': 'tests',

        :spaces_before
        TEFT/TEFTD      ' test tested'
        :spaces_after
        TEFTS/TEFT      ' test tested tests test '
        """

    def test_space_placement_change2(self):
        r"""
        'TEFT': 'test',
        'TEFTD': 'tested',
        'TEFTS': 'tests',

        :spaces_after
        TEFT/TEFTD      'test tested '
        :spaces_before
        TEFTS/TEFT      'test tested tests test'
        """

    def test_undo_after_space_placement_change1(self):
        r"""
        'TEFT': 'test',
        'TEFTD': 'tested',

        :spaces_before
        TEFT/TEFTD      ' test tested'
        :spaces_after
        *               ' test '
        *               ''
        """

    def test_undo_after_space_placement_change2(self):
        r"""
        'TEFT': 'test',
        'TEFTD': 'tested',

        :spaces_after
        TEFT/TEFTD      'test tested '
        :spaces_before
        *               'test'
        *               ''
        """

    def test_undo_after_space_placement_change3(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',
        'TEFTD': 'tested',

        :spaces_before
        KPA/TEFT/TEFTD  ' Test tested'
        :spaces_after
        *               ' Test '
        *               ''
        """

    def test_undo_after_space_placement_change4(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',
        'TEFTD': 'tested',

        :spaces_after
        KPA/TEFT/TEFTD  ' Test tested '
        :spaces_before
        *               ' Test'
        *               ''
        """

    def test_undo_with_space_placement_changes(self):
        r"""
        'TEFT': 'test',

        TEFT/TEFT/TEFT  ' test test test'
        :spaces_after
        *               ' test test '
        :spaces_before
        *               ' test'
        :spaces_after
        *               ''
        """

    def test_carry_capitalization1(self):
        r"""
        'KA*US': "{~|'^}cause",
        'TEFT': 'test',
        'P-P': '{.}',

        P-P/KA*US/TEFT  ". 'Cause test"
        """

    def test_carry_capitalization2(self):
        r"""
        "KR-GS": "{^~|\"}",

        KR-GS  '"'
        """

    def test_carry_capitalization3(self):
        r"""
        'TP*U': '{<}',
        "TK-LS": "{^}",
        'TEFT': 'test',
        'S-FBGS': '{^suffix}',

        TP*U/TEFT/S-FBGS/TK-LS/S-FBGS  ' TESTSUFFIXSUFFIX'
        """

    def test_carry_capitalization4(self):
        r"""
        'TP*U': '{<}',
        "TK-LS": "{^}",
        'TEFT': 'test',
        "S-P": "{^ ^}"

        TP*U/TEFT/S-P/TEFT  ' TEST test'
        """

    def test_carry_capitalization5(self):
        r"""
        'TP*U': '{<}',
        "TK-LS": "{^}",
        'TEFT': 'test',
        'S-FBGS': '{^suffix}',
        "S-P": "{^ ^}"

        TP*U/TEFT/S-FBGS/S-P/S-FBGS  ' TESTSUFFIX suffix'
        """

    def test_capitalize1(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',

        KPA/TEFT  ' Test'
        """

    def test_capitalize2(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',

        :start_attached
        KPA/TEFT  ' Test'
        """

    def test_capitalize3(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',

        :spaces_after
        KPA/TEFT  ' Test '
        """

    def test_capitalize4(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': 'test',

        :start_attached
        :spaces_after
        KPA/TEFT  ' Test '
        """

    def test_retro_capitalize1(self):
        r"""
        'KWEUP': 'equip',
        'RUP': '{*-|}',

        KWEUP/RUP  ' Equip'
        """

    def test_retro_capitalize2(self):
        r"""
        'KWEUP': 'equip',
        'RUP': '{*-|}',

        :spaces_after
        KWEUP/RUP  'Equip '
        """

    def test_retro_capitalize3(self):
        r"""
        'TEFT': 'tèśtîñg',
        'RUP': '{*-|}',

        TEFT/RUP  ' Tèśtîñg'
        """

    def test_retro_capitalize4(self):
        r"""
        'PRE': '{pre^}',
        'TPEUBG': 'fix',
        'RUP': '{*-|}',

        PRE/TPEUBG/RUP  ' Prefix'
        """

    def test_retro_capitalize5(self):
        r"""
        'KWEUP': 'equip',
        'RUP': '{:retro_case:cap_first_word}',

        KWEUP/RUP  ' Equip'
        """

    def test_retro_currency1(self):
        r"""
        'TPHAPB': 'notanumber',
        'R-BG': '{*($c)}',

        TPHAPB/R-BG  ' notanumber'
        """

    def test_retro_currency2(self):
        r"""
        'TPHAPB': 'notanumber',
        'R-BG': '{*($c)}',

        :spaces_after
        TPHAPB/R-BG  'notanumber '
        """

    def test_retro_currency3(self):
        r"""
        'R-BG': '{*($c)}',

        0/R-BG  ' $0'
        """

    def test_retro_currency4(self):
        r"""
        'R-BG': '{*($c)}',

        :spaces_after
        0/R-BG  '$0 '
        """

    def test_retro_currency5(self):
        r"""
        'R-BG': '{:retro_currency:$c}',

        0/R-BG  ' $0'
        """

    def test_retro_currency6(self):
        r"""
        'R-BG': '{:retro_currency:$c}',
        'THO*U': '{^},000'

        23/THO*U/R-BG  ' $23,000'
        """

    def test_retro_currency7(self):
        r"""
        'R-BG': '{:retro_currency:$c}',
        'P-P': '{^}.{^}',
        'THO*U': '{^},000'

        23/THO*U/P-P/15/R-BG  ' $23,000.15'
        """

    def test_retro_currency8(self):
        r"""
        'R-BG': '{:retro_currency:$c}',
        'P-P': '{^}.{^}',
        'TPR*UPBT': '{^},500,000'

        4/3/1/TPR*UPBT/P-P/69/R-BG  ' $431,500,000.69'
        """

    def test_retro_upper1(self):
        r"""
        'TEFT': 'test',
        '-G': '{^ing}',
        'PRE': '{pre^}',
        'R*U': '{*<}',

        TEFT/-G/R*U/PRE  " TESTING pre"
        """

    def test_retro_upper2(self):
        r"""
        'TEFT': 'test',
        '-G': '{^ing}',
        'PRE': '{pre^}',
        'R*U': '{*<}',

        TEFT/R*U/-G/PRE  " TESTing pre"
        """

    def test_retro_upper3(self):
        r"""
        'PRE': '{prefix^}',
        'WORD': 'word',
        'R*U': '{*<}',

        PRE/WORD/R*U  " PREFIXWORD"
        """

    def test_retro_upper4(self):
        r"""
        'S-G': 'something',
        'R*U': '{*<}',

        S-G/S-G/R*U/R*U  " something SOMETHING"
        """

    def test_retro_upper5(self):
        r"""
        'TEFT': 'tèśtîñg',
        'RUP': '{*<}',

        TEFT/RUP  ' TÈŚTÎÑG'
        """

    def test_retro_upper6(self):
        r"""
        'ST': 'it is',
        'RUP': '{*<}',

        ST/RUP  " it IS"
        """

    def test_retro_upper7(self):
        r"""
        'TEFT': 'test',
        "W-G": "{^ing with}",
        'RUP': '{*<}',

        TEFT/RUP/W-G  " TESTing with"
        """

    def test_retro_upper8(self):
        r"""
        'TEFT': 'test',
        "W-G": "{^ing with}",
        'RUP': '{*<}',

        TEFT/W-G/RUP  " testing WITH"
        """

    def test_retro_upper9(self):
        r"""
        'TEFT': 'test',
        "W-G": "{^ing with}",
        'RUP': '{*<}',

        TEFT/RUP/W-G/W-G  " TESTing withing with"
        """

    def test_retro_upper10(self):
        r"""
        'TEFT': 'test',
        "W": "with",
        'RUP': '{*<}',

        TEFT/RUP/W/W  " TEST with with"
        """

    def test_retro_upper11(self):
        r"""
        "PREPB": "{~|(^}",
        'TEFT': 'test',
        'RUP': '{*<}',

        PREPB/TEFT/RUP  " (TEST"
        """

    def test_retro_upper12(self):
        r"""
        'PRE': '{pre^}',
        "PREPB": "{~|(^}",
        'TEFT': 'test',
        'RUP': '{*<}',

        PRE/PREPB/TEFT/RUP  " pre(TEST"
        """

    def test_retro_upper13(self):
        r"""
        "PEUD": "pid",
        "TPAOEUL": "file",
        "H*PB": "{^-^}",
        'RUP': '{*<}',

        PEUD/RUP/H*PB/TPAOEUL  ' PID-file'
        """

    def test_retro_upper14(self):
        r"""
        "PEUD": "pid",
        "TPAOEUL": "file",
        "H*PB": "{^-^}",
        'RUP': '{*<}',

        PEUD/H*PB/TPAOEUL/RUP  ' PID-FILE'
        """

    def test_retro_upper15(self):
        r"""
        "OEU": "{^/^}",
        "T*": "{>}{&t}",
        "A*": "{>}{&a}",
        "KR*": "{>}{&c}",
        "O*": "{>}{&o}",
        "S*": "{>}{&s}",
        'RUP': '{*<}',

        T*/A*/OEU/KR*/O*/S*/RUP  ' ta/COS'
        """

    def test_retro_upper16(self):
        r"""
        "S*": "{>}{&s}",
        "T*": "{>}{&t}",
        "O*": "{>}{&o}",
        "STA*R": "{^*^}",
        "*E": "{>}{&e}",
        "*U": "{>}{&u}",
        "P*": "{>}{&p}",
        "PW*": "{>}{&b}",
        'RUP': '{*<}',

        S*/T*/O*/STA*R/*E/*U/P*/PW*/RUP  ' sto*EUPB'
        """

    def test_retro_upper17(self):
        r"""
        "*U": "{>}{&u}",
        "S*": "{>}{&s}",
        "A*": "{>}{&a}",
        "P-P": "{^.^}",
        'RUP': '{*<}',

        *U/P-P/S*/P-P/A*/P-P/RUP  ' u.s.a.'
        """

    def test_retro_upper18(self):
        r"""
        "TPAO": "foo",
        "KPATS": "{*-|}",
        "AES": "{^'s}",

        TPAO/AES/KPATS " Foo's"
        """

    def test_retro_upper19(self):
        r"""
        "TPAO": "foo",
        "KPATS": "{*<}",

        TPAO " foo"
        KPATS " FOO"
        * " foo"
        """

    def test_retro_upper20(self):
        r"""
        "TPAO": "foo",
        "KPATS": "{*<}",

        :spaces_after
        TPAO "foo "
        KPATS "FOO "
        * "foo "
        """

    def test_retro_upper21(self):
        r"""
        'TEFT': 'test',
        'RUP': '{:retro_case:upper_FIRST_word}',

        TEFT/RUP  " TEST"
        """

    def test_lower_first_char_1(self):
        r"""
        'TEFT': '{:CASE:Lower_First_Char}TEST',

        TEFT  ' tEST'
        """

    def test_upper1(self):
        r"""
        'TP*U': '{<}',
        'TEFT': 'test',
        '-G': '{^ing}',
        'PRE': '{pre^}',

        TP*U/TEFT/-G/PRE  " TESTING pre"
        """

    def test_upper2(self):
        r"""
        'TP*U': '{<}',
        'TEFT': 'test',
        '-G': '{^ing}',

        TP*U/TEFT/-G/TEFT  " TESTING test"
        """

    def test_upper3(self):
        r"""
        'TP*U': '{<}',
        'ST': 'it is',

        TP*U/ST  " IT is"
        """

    def test_upper4(self):
        r"""
        'TP*U': '{<}',
        'ST': 'it{ }is',

        TP*U/ST  " IT is"
        """

    def test_upper5(self):
        r"""
        'TP*U': '{<}',
        'TEFT': 'test',
        "W-G": "{^ing with}",

        TP*U/TEFT/W-G  " TESTING with"
        """

    def test_upper6(self):
        r"""
        'TP*U': '{<}',
        'P-': '{foo^}',
        '-S': '{^bar}',

        TP*U/P-/-S  " FOOBAR"
        """

    def test_upper7(self):
        r"""
        'TP*U': '{<}',
        'PRE': '{pre^}',
        'TEFT': 'test',
        '-G': '{^ing}',

        TP*U/PRE/TEFT/-G  " PRETESTING"
        """

    def test_upper8(self):
        r"""
        'TP*U': '{<}',
        'TEFT': 'test',
        "W-G": "{^ing with}",

        TP*U/TEFT/W-G/W-G  " TESTING withing with"
        """

    def test_upper9(self):
        r"""
        'TP*U': '{<}',
        'TEFT': 'test',
        "W": "with",

        TP*U/TEFT/W/W  " TEST with with"
        """

    def test_upper10(self):
        r"""
        'TEFT': '{:case:upper_first_word}test',
        '-G': '{^ing}',

        TEFT/-G  ' TESTING'
        """

    def test_attach_glue_and_carry_capitalize(self):
        r"""
        'PH*': '{&m}',
        'KW-GS': '{~|"}',

        PH*/KW-GS  ' m "'
        """

    def test_fingerspelling_retro_meta1(self):
        r"""
        'K': '{&c}',
        'A': '{&a}',
        'T': '{&t}',
        'UP': '{*<}',

        K/A/T  ' cat'
        UP     ' CAT'
        *      ' cat'
        """

    def test_fingerspelling_retro_meta2(self):
        r"""
        'TPH': '{>}{&n}',
        'O': '{>}{&o}',
        'UP': '{*<}',

        TPH/O/UP  ' NO'
        """

    def test_fingerspelling_retro_meta3(self):
        r"""
        'TEFT': '{>}{&n}{>}{&o}{*<}',

        TEFT  ' NO'
        """

    def test_word_1(self):
        r"""
        "KA*PS": "{MODE:CAPS}",
        "TEFT": "test",
        "-G": "{^ing}",
        'RUP': '{*-|}',

        KA*PS/TEFT/-G  ' TESTING'
        RUP            ' TESTING'
        """

    def test_word_2(self):
        r"""
        "KA*PS": "{MODE:CAPS}",
        "R*FT": "{MODE:RESET}",
        "TEFT": "test",
        "-G": "{^ing}",
        'RUL': '{*>}',

        KA*PS/TEFT/-G  ' TESTING'
        R*FT/RUL       ' tESTING'
        """

    def test_cat_burger_1(self):
        r"""
        "KAT": "cat",
        "O*PB":  "{^on}",
        "PWURG": "{*-|}{^burg}",
        "*ER": "{^er}",
        "PWURG/*ER": "burger",

        KAT/O*PB/PWURG  ' Cattonburg'
        *ER             ' catton burger'
        """

    def test_cat_burger_2(self):
        r"""
        "KAT": "cat",
        "O*PB":  "{^on}",
        "PWURG": "{*-|}{^burg}",
        "*ER": "{^er}",

        KAT/O*PB/PWURG  ' Cattonburg'
        *ER             ' Cattonburger'
        """

    def test_mirrored_capitalize(self):
        r"""
        'KPA': '{}{-|}',
        'TEFT': '{~|na^}{~|no^}{~|wri^}mo'

        KPA/TEFT  ' NaNoWriMo'
        """

    def test_mode_1a(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",

        :spaces_after
        TEFT   'test '
        PHO*D  'test '
        TEFT   'testtest'
        TEFT   'testtestTest'
        """

    def test_mode_1b(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",

        :spaces_before
        TEFT   ' test'
        PHO*D  ' test'
        TEFT   ' testtest'
        TEFT   ' testtestTest'
        """

    def test_mode_2a(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",
        "TKPWHRAOU": "{&g}{&l}{&u}{&e}"

        :spaces_after
        PHO*D  ''
        TEFT       'test'
        TKPWHRAOU  'testGlue'
        """

    def test_mode_2b(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",
        "TKPWHRAOU": "{&g}{&l}{&u}{&e}"

        :spaces_before
        PHO*D  ''
        TEFT       'test'
        TKPWHRAOU  'testGlue'
        """

    def test_mode_3a(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",
        "TKPWHRAOU": "{&g}{&l}{&u}{&e}"

        :spaces_after
        PHO*D  ''
        TKPWHRAOU  'glue'
        TEFT       'glueTest'
        """

    def test_mode_3b(self):
        r"""
        "PHO*D": "{MODE:CAMEL}",
        "TEFT": "test",
        "TKPWHRAOU": "{&g}{&l}{&u}{&e}"

        :spaces_before
        PHO*D  ''
        TKPWHRAOU  'glue'
        TEFT       'glueTest'
        """

    def test_mode_4(self):
        r"""
        "PHO*D": "{:mode:Camel}",
        "TEFT": "test",
        "TKPWHRAOU": "{&g}{&l}{&u}{&e}"

        PHO*D  ''
        TKPWHRAOU  'glue'
        TEFT       'glueTest'
        """

    def test_fingerspelling_1a(self):
        r"""
        'K': '{&c}',
        'A': '{&a}',
        'T': '{&t}',

        :spaces_after
        K/A/T  'cat '
        """

    def test_fingerspelling_1b(self):
        r"""
        'K': '{&c}',
        'A': '{&a}',
        'T': '{&t}',

        K/A/T  ' cat'
        """

    def test_fingerspelling_2a(self):
        r"""
        'K': '{-|}{&c}',
        'A': '{-|}{&a}',
        'T': '{-|}{&t}',

        :spaces_after
        K/A/T  'CAT '
        """

    def test_fingerspelling_2b(self):
        r"""
        'K': '{-|}{&c}',
        'A': '{-|}{&a}',
        'T': '{-|}{&t}',

        K/A/T  ' CAT'
        """

    def test_numbers_1a(self):
        r"""
        '': ''

        :spaces_after
        1-9  '19 '
        -79  '1979 '
        *    '19 '
        *    ''
        """

    def test_numbers_1b(self):
        r"""
        '': ''

        1-9  ' 19'
        -79  ' 1979'
        *    ' 19'
        *    ''
        """

    def test_raw_1a(self):
        r"""
        '': ''

        :spaces_after
        RAU   'RAU '
        TEGT  'RAU TEGT '
        *     'RAU '
        *     ''
        """

    def test_raw_1b(self):
        r"""
        '': ''

        RAU   ' RAU'
        TEGT  ' RAU TEGT'
        *     ' RAU'
        *     ''
        """

    def test_bug287(self):
        r"""
        "*6": "a",
        "R*S": "{*+}",

        *6   ' a'
        R*S  ' a a'
        """

    def test_bug470(self):
        r"""
        "0EU8": "80",
        "R*S": "{*+}",

        0EU8  ' 80'
        R*S   ' 8080'
        """

    def test_bug851(self):
        r"""
        "KUPBTS": "countries",
        "R-R": "{^}{#Return}{^}{-|}",

        :spaces_after
        KUPBTS "countries "
        R-R "countries"
        """

    def test_bug849_1(self):
        r"""
        "-P": ".",
        "*P": "{*}",
        "TKAOU": "due",
        "TKAO*U": "dew",

        TKAOU  ' due'
        -P      ' due .'
        *P     ' dew'
        """

    def test_bug849_2(self):
        r"""
        "KPA*": "{^}{-|}",
        "*P": "{*}",
        "TKAOU": "due",
        "TKAO*U": "dew",

        KPA*   ''
        TKAOU  'Due'
        *P     'Dew'
        """

    def test_carry_upper_spacing1(self):
        r"""
        "TEFT": "{<}test",
        "-G": "{^ing}",
        "S-P": "{^ ^}",
        "S-G": "something",

        TEFT/-G  ' TESTING'
        S-P      ' TESTING '
        S-G      ' TESTING something'
        """

    def test_carry_upper_spacing2(self):
        r"""
        "TEFT": "{<}test",
        "W-G": "{^ing with}",
        "S-G": "something",

        TEFT/W-G  ' TESTING with'
        S-G      ' TESTING with something'
        """

    def test_carry_upper_spacing3(self):
        r"""
        "TEFT": "{<}test",
        "-G": "{^ing}",
        "R-R": "{^\n^}",
        "S-G": "something",

        TEFT/-G  ' TESTING'
        R-R      ' TESTING\n'
        S-G      ' TESTING\nsomething'
        """

    def test_carry_upper_spacing4(self):
        r"""
        "TEFT": "{<}test",
        "W-G": "{^ing\twith}",

        TEFT/W-G  ' TESTING\twith'
        W-G      ' TESTING\twithing\twith'
        """

    def test_carry_upper_spacing5(self):
        r"""
        "TEFT": "{<}test",
        "-G": "{^ing}",
        "TA*B": "{^\t^}",
        "S-G": "something",

        TEFT/-G  ' TESTING'
        TA*B     ' TESTING\t'
        S-G      ' TESTING\tsomething'
        """

    def test_carry_upper_spacing6(self):
        r"""
        "TEFT": "{<}test",
        "W-G": "{^ing\nwith}",

        TEFT/W-G  ' TESTING\nwith'
        W-G       ' TESTING\nwithing\nwith'
        """

    def test_bug961(self):
        r"""
        "PHAED": "made",
        "KWREUS": "{^ies}",

        :spaces_after
        PHAED/KWREUS  'madies '
        *             'made '
        """

    def test_bug966_1(self):
        r"""
        "-G": "{^ing}",
        "KPA*L": "{<}",
        "TKPWAEUPL": "game",

        KPA*L/TKPWAEUPLG  ' GAMING'
        """

    def test_bug966_2(self):
        r"""
        "-G": "{^ing}",
        "KPA*L": "{<}",
        "TKPWAEUPL": "game",

        KPA*L/TKPWAEUPL/-G/*  ' GAME'
        """

    def test_bug980(self):
        r"""
        "PUFRPB": "punch",
        "HRAOEUPB": "line",
        "PUFRPB/HRAOEUPB": "punchline",
        "-S": "{^s}",
        "AFPS": "{*?}",

        PUFRPB/HRAOEUPBS  ' punchlines'
        AFPS              ' punch lines'
        """

    def test_not_a_macro_1(self):
        r"""
        "TEFT": "=",

        TEFT  ' ='
        """

    def test_not_a_macro_2(self):
        r"""
        "TEFT": "==",

        TEFT  ' =='
        """

    def test_not_a_macro_3(self):
        r"""
        "TEFT": "=>",

        TEFT  ' =>'
        """

    def test_not_a_macro_4(self):
        r"""
        "TEFT": "== 0",

        TEFT  ' == 0'
        """

    def test_not_a_macro_5(self):
        r"""
        "TEFT": "=not a macro",

        TEFT  ' =not a macro'
        """

    def test_not_a_macro_6(self):
        r"""
        "TEFT": "=not a macro:",

        TEFT  ' =not a macro:'
        """

    def test_bad_macro_1(self):
        r"""
        "TEFT": "=not_a_known_macro",

        TEFT  raise KeyError
        """

    def test_bad_macro_2(self):
        # Builtin macros not taking parameters.
        r"""
        "1": "=undo:param",
        "2": "=repeat_last_stroke:param",
        "3": "=retrospective_toggle_asterisk:param",
        "4": "=retrospective_delete_space:param",
        "5": "=retrospective_insert_space:param",

        1  raise AssertionError
        2  raise AssertionError
        3  raise AssertionError
        4  raise AssertionError
        5  raise AssertionError
        """

    def test_valid_macro_1(self):
        r"""
        "1": "=undo",
        "2": "=undo:",

        1   ""
        2   ""
        """

    def test_valid_macro_2(self, monkeypatch):
        r"""
        "TEFT": "=macro: params with spaces   ",

        TEFT   ""
        """

        def macro(translator, stroke, cmdline):
            assert cmdline == " params with spaces   "

        registry = Registry()
        registry.register_plugin("macro", "macro", macro)
        monkeypatch.setattr("plover.translation.registry", registry)

    def test_valid_macro_3(self, monkeypatch):
        r"""
        "TEFT": "=macro",

        TEFT   ""
        """

        def macro(translator, stroke, cmdline):
            assert cmdline == ""

        registry = Registry()
        registry.register_plugin("macro", "macro", macro)
        monkeypatch.setattr("plover.translation.registry", registry)

    def test_meta_attach_default(self):
        r"""
        'TEFT': 'test',
        'AT': '{:attach:attach}',

        TEFT/AT/TEFT  ' testattachtest'
        """

    def test_meta_attach_infix(self):
        r"""
        'TEFT': 'test',
        'AT': '{:attach:^attach^}',

        TEFT/AT/TEFT  ' testattachtest'
        """

    def test_meta_attach_prefix(self):
        r"""
        'TEFT': 'test',
        'AT': '{:attach:attach^}',

        TEFT/AT/TEFT  ' test attachtest'
        """

    def test_meta_attach_suffix(self):
        r"""
        'TEFT': 'test',
        'AT': '{:attach:^attach}',

        TEFT/AT/TEFT  ' testattach test'
        """

    def test_prefix_strokes(self):
        r"""
        "/S": "{prefix^}",
        "S": "{^suffix}",
        "O": "{O'^}{$}",

        S/S/O/S/S  " prefixsuffix O'prefixsuffix"
        """

    def test_melani_implicit_hyphens(self, with_melani_system):
        r"""
        :system Melani
        "15/SE/COhro": "XV secolo",
        "16/SE/COhro": "XVI secolo",

        15/SE/COhro/16/SE/COhro  " XV secolo XVI secolo"
        """

    def test_conditionals_1(self):
        r"""
        "*": "=undo",
        "S-": "{=(?i)t/true/false}",
        "TP-": "FALSE",
        "T-": "TRUE",

        S-   ' false'
        TP-  ' false FALSE'
        *    ' false'
        T-   ' true TRUE'
        *    ' false'
        S-   ' false false'
        TP-  ' false false FALSE'
        *    ' false false'
        T-   ' true true TRUE'
        """

    def test_conditionals_2(self):
        r"""
        "1": "{=(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/agli/ai}",
        "2": "oc{^}chi",
        "3": "dei",
        "4": "sti{^}vali",

        1  ' ai'
        2  ' agli occhi'
        1  ' agli occhi ai'
        3  ' agli occhi agli dei'
        1  ' agli occhi agli dei ai'
        4  ' agli occhi agli dei agli stivali'
        """

    def test_conditionals_3(self):
        r"""
        "1": "{:if_next_matches:(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/agli/ai}",
        "2": "chi",
        "3": "oc{^}chi",
        "4": "dei",
        "5": "sti{^}vali",

        :spaces_after
        2  'chi '
        1  'chi ai '
        3  'chi agli occhi '
        1  'chi agli occhi ai '
        4  'chi agli occhi agli dei '
        1  'chi agli occhi agli dei ai '
        5  'chi agli occhi agli dei agli stivali '
        """

    def test_conditionals_4(self):
        r"""
        "1": "{=(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/agli/ai}",
        "2": "{=(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/cogli/coi}",
        "3": "ci",

        1/2/1/3  ' ai cogli ai ci'
        """

    def test_conditionals_5(self):
        r"""
        "1": "{=(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/agli/ai}",
        "2": "{=(?i)([8aeiouxy]|11|dei|gn|ps|s[bcdfglmnpqrtv]|z)/cogli/coi}",
        "3": "ci",

        :spaces_after
        1/2/1/3  'ai cogli ai ci '
        """

    def test_conditionals_6(self):
        r"""
        "*": "=undo",
        "S-": r'{=(?i)tr\/ue/tr\/ue/fa\\lse}',
        "TP-": r'FA\LSE',
        "T-": 'TR/UE',

        S-   r' fa\lse'
        TP-  r' fa\lse FA\LSE'
        *    r' fa\lse'
        T-   r' tr/ue TR/UE'
        *    r' fa\lse'
        S-   r' fa\lse fa\lse'
        TP-  r' fa\lse fa\lse FA\LSE'
        *    r' fa\lse fa\lse'
        T-   r' tr/ue tr/ue TR/UE'
        """

    def test_bug_1448_1(self, with_korean_system):
        # Translator tries to represent a previous stroke
        # from a different system using the current one.
        #
        # This test would throw a `ValueError` exception
        # (invalid keys mask) in the translator when
        # trying to represent the `ㅎㅁㄱㅈㄴ-ㄴㅅㅈㅁ`
        # stroke with the `English Stenotype` system.
        r"""
        'TEFT': 'test'
        'TEFT/-G': 'testing'

        :system 'Korean Modern C'
        ㅎㅁㄱㅈㄴ-ㄴㅅㅈㅁ  ' ㅎㅁㄱㅈㄴ-ㄴㅅㅈㅁ'
        :system 'English Stenotype'
        TEFT  ' ㅎㅁㄱㅈㄴ-ㄴㅅㅈㅁ test'
        """

    def test_bug_1448_2(self, with_korean_system):
        # Translator tries to represent a previous stroke
        # from a different system using the current one.
        #
        # This test would trigger a translator lookup for
        # `#STKP/-G` because `#STKP` (English Stenotype)
        # and `12345` (Korean Modern C) have the same keys
        # mask.
        r"""
        '#STKP/-G': 'game over'

        :system 'Korean Modern C'
        12345  ' 12345'
        :system 'English Stenotype'
        -G  ' 12345 -G'
        """



import json
from collections import defaultdict, OrderedDict

from plover import log


class Keymap:
    def __init__(self, keys, actions):
        # List of supported actions.
        self._actions = OrderedDict((action, n) for n, action in enumerate(actions))
        self._actions["no-op"] = len(self._actions)
        # List of supported keys.
        self._keys = OrderedDict((key, n) for n, key in enumerate(keys))
        # action -> keys
        self._mappings = {}
        # key -> action
        self._bindings = {}

    def get_keys(self):
        return self._keys.keys()

    def get_actions(self):
        return self._actions.keys()

    def set_bindings(self, bindings):
        # Set from:
        # { key1: action1, key2: action1, ... keyn: actionn }
        mappings = defaultdict(list)
        for key, action in dict(bindings).items():
            mappings[action].append(key)
        self.set_mappings(mappings)

    def set_mappings(self, mappings):
        # When setting from a string, assume a list of mappings:
        # [[action1, [key1, key2]], [action2, [key3]], ...]
        if isinstance(mappings, str):
            mappings = json.loads(mappings)
        mappings = dict(mappings)
        # Set from:
        # { action1: [key1, key2], ... actionn: [keyn] }
        self._mappings = OrderedDict()
        self._bindings = {}
        bound_keys = defaultdict(list)
        errors = []
        for action in self._actions:
            key_list = mappings.get(action)
            if not key_list:
                # Not an issue if 'no-op' is not mapped...
                if action != "no-op":
                    errors.append("action %s is not bound" % action)
                # Add dummy mapping for each missing action
                # so it's shown in the configurator.
                self._mappings[action] = ()
                continue
            if isinstance(key_list, str):
                key_list = (key_list,)
            valid_key_list = []
            for key in key_list:
                if key not in self._keys:
                    errors.append("invalid key %s bound to action %s" % (key, action))
                    continue
                valid_key_list.append(key)
                bound_keys[key].append(action)
                self._bindings[key] = action
            self._mappings[action] = tuple(sorted(valid_key_list, key=self._keys.get))
        for action in set(mappings) - set(self._actions):
            key_list = mappings.get(action)
            if isinstance(key_list, str):
                key_list = (key_list,)
            errors.append(
                "invalid action %s mapped to key(s) %s" % (action, " ".join(key_list))
            )
        for key, action_list in bound_keys.items():
            if len(action_list) > 1:
                errors.append(
                    "key %s is bound multiple times: %s" % (key, str(action_list))
                )
        if len(errors) > 0:
            log.warning(
                "Keymap is invalid, behavior undefined:\n\n- " + "\n- ".join(errors)
            )

    def get_bindings(self):
        return self._bindings

    def get_mappings(self):
        return self._mappings

    def get_action(self, key, default=None):
        return self._bindings.get(key, default)

    def keys_to_actions(self, key_list):
        action_list = []
        for key in key_list:
            assert key in self._keys, "'%s' not in %s" % (key, self._keys)
            action = self._bindings[key]
            if "no-op" != action:
                action_list.append(action)
        return action_list

    def keys(self):
        return self._mappings.keys()

    def values(self):
        return self._mappings.values()

    def __len__(self):
        return len(self._mappings)

    def __getitem__(self, key):
        return self._mappings[key]

    def __setitem__(self, action, key_list):
        assert action in self._actions
        if isinstance(key_list, str):
            key_list = (key_list,)
        # Delete previous bindings.
        if action in self._mappings:
            for old_key in self._mappings[action]:
                if old_key in self._bindings:
                    del self._bindings[old_key]
        errors = []
        valid_key_list = []
        for key in key_list:
            if key not in self._keys:
                errors.append("invalid key %s bound to action %s" % (key, action))
                continue
            if key in self._bindings:
                errors.append(
                    "key %s is already bound to: %s" % (key, self._bindings[key])
                )
                continue
            valid_key_list.append(key)
            self._bindings[key] = action
        self._mappings[action] = tuple(sorted(valid_key_list, key=self._keys.get))
        if len(errors) > 0:
            log.warning(
                "Keymap is invalid, behavior undefined:\n\n- " + "\n- ".join(errors)
            )

    def __iter__(self):
        return iter(self._mappings)

    def __eq__(self, other):
        return self.get_mappings() == other.get_mappings()

    def __str__(self):
        # Use the more compact list of mappings format:
        # [[action1, [key1, key2]], [action2, [key3]], ...]
        return json.dumps(list(self._mappings.items()))



# Copyright (c) 2016 Ted Morin
# See LICENSE.txt for details.

"""Thread-based monitoring of a ProCAT stenotype machine."""

import binascii

from plover import log
from plover.machine.base import SerialStenotypeBase


# ProCAT machines send 4 bytes per stroke, with the last byte only consisting of
# FF. So we need only look at the first 3 bytes to see our steno. The leading
# bit is 0.
# fmt: off
STENO_KEY_CHART = (None, "#", "S-", "T-", "K-", "P-", "W-", "H-",
                   "R-", "A-", "O-", "*", "-E", "-U", "-F", "-R",
                   "-P", "-B", "-L", "-G", "-T", "-S", "-D", "-Z",
                   )
# fmt: on

BYTES_PER_STROKE = 4


class ProCat(SerialStenotypeBase):
    """Interface for ProCAT machines."""

    KEYS_LAYOUT = """
        #  #  #  #  #  #  #  #  #  #
        S- T- P- H- * -F -P -L -T -D
        S- K- W- R- * -R -B -G -S -Z
               A- O- -E -U
    """
    KEYMAP_MACHINE_TYPE = "TX Bolt"

    def run(self):
        """Overrides base class run method. Do not call directly."""
        self._ready()
        for packet in self._iter_packets(BYTES_PER_STROKE):
            if (packet[0] & 0x80) or packet[3] != 0xFF:
                log.error("discarding invalid packet: %s", binascii.hexlify(packet))
                continue
            self._notify_keys(self.process_steno_packet(packet))

    @staticmethod
    def process_steno_packet(raw):
        # Raw packet has 4 bytes, we only care about the first 3
        steno_keys = []
        for i, b in enumerate(raw[:3]):
            for j in range(0, 8):
                if b & 0x80 >> j:
                    key = STENO_KEY_CHART[i * 8 + j]
                    steno_keys.append(key)
        return steno_keys



from collections import defaultdict
from contextlib import contextmanager
import ast
import functools
import inspect
import os

import pytest

from plover.registry import registry
from plover.resource import ASSET_SCHEME
from plover.steno import normalize_steno
from plover.steno_dictionary import StenoDictionary

from .dict import make_dict
from .parametrize import parametrize


class _DictionaryTests:
    @staticmethod
    def make_dict(contents):
        return contents

    @contextmanager
    def tmp_dict(self, tmp_path, contents):
        contents = self.make_dict(contents)
        with make_dict(tmp_path, contents, extension=self.DICT_EXTENSION) as dict_path:
            yield dict_path

    @contextmanager
    def sample_dict(self, tmp_path):
        with self.tmp_dict(tmp_path, self.DICT_SAMPLE) as dict_path:
            yield dict_path

    @staticmethod
    def parse_entries(entries):
        return {
            normalize_steno(k): v
            for k, v in ast.literal_eval("{" + entries + "}").items()
        }

    def test_readonly_writable_file(self, tmp_path):
        """
        Writable file: match class read-only attribute.
        """
        with self.sample_dict(tmp_path) as dict_path:
            d = self.DICT_CLASS.load(str(dict_path))
            assert d.readonly == self.DICT_CLASS.readonly

    def test_readonly_readonly_file(self, tmp_path):
        """
        Read-only file: read-only dictionary.
        """
        with self.sample_dict(tmp_path) as dict_path:
            dict_path.chmod(0o440)
            try:
                d = self.DICT_CLASS.load(str(dict_path))
                assert d.readonly
            finally:
                # Deleting the file will fail on Windows
                # if we don't restore write permission.
                dict_path.chmod(0o660)

    def test_readonly_asset(self, tmp_path, monkeypatch):
        """
        Assets are always read-only.
        """
        with self.sample_dict(tmp_path) as dict_path:
            fake_asset = ASSET_SCHEME + "fake:" + dict_path.name

            def fake_asset_only(r, v=None):
                assert r.startswith(ASSET_SCHEME + "fake:")
                return v

            monkeypatch.setattr(
                "plover.resource._asset_filename",
                functools.partial(fake_asset_only, v=str(dict_path)),
            )
            d = self.DICT_CLASS.load(fake_asset)
        assert d.readonly

    VALID_KEY = ("TEFT", "-G")

    @pytest.mark.parametrize(
        "method_name, args",
        (
            ("__delitem__", (VALID_KEY,)),
            ("__setitem__", (VALID_KEY, "pouet!")),
            ("clear", ()),
            ("save", ()),
            ("update", ()),
        ),
    )
    def test_readonly_no_change_allowed(self, tmp_path, method_name, args):
        """
        Don't allow changing a read-only dictionary.
        """
        with self.sample_dict(tmp_path) as dict_path:
            d = self.DICT_CLASS.load(str(dict_path))
        d.readonly = True
        method = getattr(d, method_name)
        with pytest.raises(AssertionError):
            method(*args)

    def _test_entrypoint(self):
        """
        Check a corresponding `plover.dictionary` entrypoint exists.
        """
        plugin = registry.get_plugin("dictionary", self.DICT_EXTENSION)
        assert plugin.obj == self.DICT_CLASS

    DUMMY = object()
    MISSING_KEY = "ceci n'est pas une clef"
    MISSING_TRANSLATION = "ceci n'est pas une translation"

    def _test_load(self, tmp_path, contents, expected):
        """
        Test `load` implementation.
        """
        with self.tmp_dict(tmp_path, contents) as dict_path:
            expected_timestamp = dict_path.stat().st_mtime
            if inspect.isclass(expected):
                # Expect an exception.
                with pytest.raises(expected):
                    self.DICT_CLASS.load(str(dict_path))
                return
            # Except a successful load.
            d = self.DICT_CLASS.load(str(dict_path))
        # Parse entries:
        entries = self.parse_entries(expected)
        # - expected: must be present
        expected_entries = {k: v for k, v in entries.items() if v is not None}
        # - unexpected: must not be present
        unexpected_entries = {k for k, v in entries.items() if v is None}
        # Basic checks.
        assert d.readonly == self.DICT_CLASS.readonly
        assert d.timestamp == expected_timestamp
        if self.DICT_SUPPORT_SEQUENCE_METHODS:
            assert sorted(d.items()) == sorted(expected_entries.items())
            assert sorted(d) == sorted(expected_entries)
            assert len(d) == len(expected_entries)
        else:
            assert tuple(d.items()) == ()
            assert tuple(d) == ()
            assert len(d) == 0
        for k, v in expected_entries.items():
            assert k in d
            assert d[k] == v
            assert d.get(k) == v
            assert d.get(k, self.DUMMY) == v
        for k in unexpected_entries:
            assert k not in d
            with pytest.raises(KeyError):
                d[k]
            assert d.get(k) == None
            assert d.get(k, self.DUMMY) == self.DUMMY
        assert self.MISSING_KEY not in d
        assert d.get(self.MISSING_KEY, self.DUMMY) is self.DUMMY
        assert d.get(self.MISSING_KEY) == None
        with pytest.raises(KeyError):
            d[self.MISSING_KEY]
        # Longest key check.
        expected_longest_key = functools.reduce(
            max, (len(k) for k in expected_entries), 0
        )
        assert d.longest_key == expected_longest_key
        # Reverse lookup checks.
        expected_reverse = defaultdict(set)
        if self.DICT_SUPPORT_REVERSE_LOOKUP:
            for k, v in expected_entries.items():
                expected_reverse[v].add(k)
        for v, key_set in expected_reverse.items():
            assert d.reverse_lookup(v) == key_set
        assert d.reverse_lookup(self.MISSING_TRANSLATION) == set()
        # Case reverse lookup checks.
        expected_casereverse = defaultdict(set)
        if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
            for v in expected_entries.values():
                expected_casereverse[v.lower()].add(v)
        for v, value_set in expected_casereverse.items():
            assert d.casereverse_lookup(v) == value_set
        assert d.casereverse_lookup(self.MISSING_TRANSLATION) == set()


class _ReadOnlyDictionaryTests:
    def test_readonly_no_create_allowed(self, tmp_path):
        """
        Don't allow creating a read-only dictionary.
        """
        with self.sample_dict(tmp_path) as dict_path:
            with pytest.raises(ValueError):
                self.DICT_CLASS.create(str(dict_path))


_TEST_DICTIONARY_UPDATE_DICT = {
    ("S-G",): "something",
    ("SPH-G",): "something",
    ("SPH*G",): "Something",
    ("SPH", "THEUPBG"): "something",
}
_TEST_DICTIONARY_UPDATE_STENODICT = StenoDictionary()
_TEST_DICTIONARY_UPDATE_STENODICT.update(_TEST_DICTIONARY_UPDATE_DICT)


class _WritableDictionaryTests:
    def test_longest_key(self):
        """
        Check `longest_key` support.
        """
        assert self.DICT_SUPPORT_SEQUENCE_METHODS
        d = self.DICT_CLASS()
        assert d.longest_key == 0
        d[("S",)] = "a"
        assert d.longest_key == 1
        d[("S", "S", "S", "S")] = "b"
        assert d.longest_key == 4
        d[("S", "S")] = "c"
        assert d.longest_key == 4
        assert d[("S", "S")] == "c"
        del d[("S", "S", "S", "S")]
        assert d.longest_key == 2
        del d[("S",)]
        assert d.longest_key == 2
        if self.DICT_SUPPORT_REVERSE_LOOKUP:
            assert d.reverse_lookup("c") == {("S", "S")}
        else:
            assert d.reverse_lookup("c") == set()
        if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
            assert d.casereverse_lookup("c") == {"c"}
        else:
            assert d.casereverse_lookup("c") == set()
        d.clear()
        assert d.longest_key == 0
        assert d.reverse_lookup("c") == set()
        assert d.casereverse_lookup("c") == set()
        d[("S", "S")] = "c"
        assert d.longest_key == 2

    def test_casereverse_del(self):
        """
        Check deletion correctly updates `casereverse_lookup` data.
        """
        d = self.DICT_CLASS()
        d[("S-G",)] = "something"
        d[("SPH-G",)] = "something"
        if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
            assert d.casereverse_lookup("something") == {"something"}
        else:
            assert d.casereverse_lookup("something") == set()
        del d[("S-G",)]
        if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
            assert d.casereverse_lookup("something") == {"something"}
        else:
            assert d.casereverse_lookup("something") == set()
        del d[("SPH-G",)]
        assert d.casereverse_lookup("something") == set()

    @parametrize(
        (
            lambda: (dict(_TEST_DICTIONARY_UPDATE_DICT), False, True),
            lambda: (dict(_TEST_DICTIONARY_UPDATE_DICT), False, False),
            lambda: (list(_TEST_DICTIONARY_UPDATE_DICT.items()), False, True),
            lambda: (list(_TEST_DICTIONARY_UPDATE_DICT.items()), False, False),
            lambda: (list(_TEST_DICTIONARY_UPDATE_DICT.items()), True, True),
            lambda: (list(_TEST_DICTIONARY_UPDATE_DICT.items()), True, False),
            lambda: (_TEST_DICTIONARY_UPDATE_STENODICT, False, True),
            lambda: (_TEST_DICTIONARY_UPDATE_STENODICT, False, False),
        )
    )
    def test_update(self, update_from, use_iter, start_empty):
        """
        Check `update` does the right thing, including consuming iterators once.
        """
        d = self.DICT_CLASS()
        if not start_empty:
            d.update(
                {
                    ("SPH*G",): "not something",
                    ("STHEUPBG",): "something",
                    ("EF", "REU", "TH*EUPBG"): "everything",
                }
            )
            assert d[("STHEUPBG",)] == "something"
            assert d[("EF", "REU", "TH*EUPBG")] == "everything"
            if self.DICT_SUPPORT_REVERSE_LOOKUP:
                assert d.reverse_lookup("not something") == {("SPH*G",)}
            else:
                assert d.reverse_lookup("not something") == set()
            if self.DICT_SUPPORT_REVERSE_LOOKUP:
                assert d.casereverse_lookup("not something") == {"not something"}
            else:
                assert d.casereverse_lookup("not something") == set()
            assert d.longest_key == 3
        if use_iter:
            update_from = iter(update_from)
        d.update(update_from)
        assert d[("S-G",)] == "something"
        assert d[("SPH-G",)] == "something"
        assert d[("SPH*G",)] == "Something"
        assert d[("SPH", "THEUPBG")] == "something"
        if not start_empty:
            assert d[("STHEUPBG",)] == "something"
            assert d[("EF", "REU", "TH*EUPBG")] == "everything"
            assert d.reverse_lookup("not something") == set()
            if self.DICT_SUPPORT_REVERSE_LOOKUP:
                assert d.reverse_lookup("something") == {
                    ("STHEUPBG",),
                    ("S-G",),
                    ("SPH-G",),
                    ("SPH", "THEUPBG"),
                }
            else:
                assert d.reverse_lookup("something") == set()
            if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
                assert d.casereverse_lookup("something") == {"something", "Something"}
            else:
                assert d.casereverse_lookup("something") == set()
            assert d.longest_key == 3
        else:
            if self.DICT_SUPPORT_REVERSE_LOOKUP:
                assert d.reverse_lookup("something") == {
                    ("S-G",),
                    ("SPH-G",),
                    ("SPH", "THEUPBG"),
                }
            else:
                assert d.reverse_lookup("something") == set()
            if self.DICT_SUPPORT_CASEREVERSE_LOOKUP:
                assert d.casereverse_lookup("something") == {"something", "Something"}
            else:
                assert d.casereverse_lookup("something") == set()
            assert d.longest_key == 2

    INVALID_CONTENTS = b"ceci n'est pas un dictionaire"

    def _test_save(self, tmp_path, entries, expected):
        """
        Test `save` implementation.
        """
        dict_entries = self.parse_entries(entries)
        with self.tmp_dict(tmp_path, self.INVALID_CONTENTS) as dict_path:
            st = dict_path.stat()
            old_timestamp = st.st_mtime - 1
            os.utime(str(dict_path), (st.st_atime, old_timestamp))
            # Create...
            d = self.DICT_CLASS.create(str(dict_path))
            # ...must not change the target file...
            assert d.timestamp == 0
            assert dict_path.read_bytes() == self.INVALID_CONTENTS
            # ...even on update...
            d.update(dict_entries)
            assert d.timestamp == 0
            assert dict_path.read_bytes() == self.INVALID_CONTENTS
            # ...until save is called.
            d.save()
            assert d.timestamp > old_timestamp
            if expected is None:
                assert dict_path.read_bytes() != self.INVALID_CONTENTS
            else:
                assert dict_path.read_bytes() == expected
            d = self.DICT_CLASS.load(str(dict_path))
            assert sorted(d.items()) == sorted(dict_entries.items())


def _wrap_method(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        return method(*args, **kwargs)

    wrapper.__signature__ = inspect.signature(method)
    return wrapper


def dictionary_test(cls):
    """
    Torture tests for dictionary implementations.

    Usage:

    ```python

    @dictionary_test
    class TestMyDictionaryClass:

        # Mandatory: implementation class.
        DICT_CLASS = MyDictionaryClass

        # Mandatory: extension.
        DICT_EXTENSION = 'json'

        # Mandatory: valid sample dictionary contents.
        DICT_SAMPLE = b'{}'

        # Optional: if `True`, will check the implementation class
        # is registered as a valid `plover.dictionary` entrypoint.
        DICT_REGISTERED = True

        # Optional: if `False`, `len`, `__iter__`, and `items`
        # are not supported, and must respectively return:
        # 0 and empty sequences.
        # Note: only supported for read-only implementations,
        # writable implementations must support all sequence
        # methods.
        DICT_SUPPORT_SEQUENCE_METHODS = False

        # Optional: if `False`, then `reverse_lookup` is not
        # supported, and must return an empty set.
        DICT_SUPPORT_REVERSE_LOOKUP = False

        # Optional: if `False`, then `casereverse_lookup` is not
        # supported, and must return an empty set.
        # Note: if supported, then `DICT_SUPPORT_REVERSE_LOOKUP`
        # must be too.
        DICT_SUPPORT_CASEREVERSE_LOOKUP = False

        # Optional: load tests.
        DICT_LOAD_TESTS = (
            lambda: (
                # Dictionary file contents.
                b'''
                ''',
                # Expected entries, in Python-dictionary like format.
                '''
                "TEFT": 'test',
                'TEFT/-G': "testing",
                '''
            ),
        )

        # Optional: save tests.
        # Note: only for writable implementations.
        DICT_SAVE_TESTS = (
            lambda: (
                # Initial entries, in Python-dictionary like format.
                '''
                "TEFT": 'test',
                'TEFT/-G': "testing",
                '''
                # Expected saved dictionary contents, or `None`
                # to skip the byte-for-byte test of the resulting
                # file contents.
                b'''
                ''',
            ),
        )

        # Optional: if implemented, will be called to format
        # the contents before saving to a dictionary file,
        # including for load and save tests.
        @staticmethod
        def make_dict(self, contents):
            return contents

    """

    DICT_SUPPORT_SEQUENCE_METHODS = getattr(cls, "DICT_SUPPORT_SEQUENCE_METHODS", True)
    DICT_SUPPORT_REVERSE_LOOKUP = getattr(cls, "DICT_SUPPORT_REVERSE_LOOKUP", True)
    DICT_SUPPORT_CASEREVERSE_LOOKUP = getattr(
        cls, "DICT_SUPPORT_CASEREVERSE_LOOKUP", DICT_SUPPORT_REVERSE_LOOKUP
    )
    assert DICT_SUPPORT_REVERSE_LOOKUP or not DICT_SUPPORT_CASEREVERSE_LOOKUP

    base_classes = [cls, _DictionaryTests]
    if cls.DICT_CLASS.readonly:
        base_classes.append(_ReadOnlyDictionaryTests)
    else:
        base_classes.append(_WritableDictionaryTests)

    class_dict = {
        "DICT_SUPPORT_SEQUENCE_METHODS": DICT_SUPPORT_SEQUENCE_METHODS,
        "DICT_SUPPORT_REVERSE_LOOKUP": DICT_SUPPORT_REVERSE_LOOKUP,
        "DICT_SUPPORT_CASEREVERSE_LOOKUP": DICT_SUPPORT_CASEREVERSE_LOOKUP,
    }

    if getattr(cls, "DICT_REGISTERED", False):
        class_dict["test_entrypoint"] = _wrap_method(_DictionaryTests._test_entrypoint)

    if hasattr(cls, "DICT_LOAD_TESTS"):
        class_dict["test_load"] = parametrize(cls.DICT_LOAD_TESTS, arity=2)(
            _wrap_method(_DictionaryTests._test_load)
        )

    if hasattr(cls, "DICT_SAVE_TESTS"):
        assert not cls.DICT_CLASS.readonly
        class_dict["test_save"] = parametrize(cls.DICT_SAVE_TESTS, arity=2)(
            _wrap_method(_WritableDictionaryTests._test_save)
        )

    return type(cls.__name__, tuple(base_classes), class_dict)



import ast
import functools
import inspect
import operator
import re
import shlex
import textwrap

from plover import system
from plover.formatting import Formatter
from plover.steno import normalize_steno
from plover.steno_dictionary import StenoDictionary
from plover.translation import Translator

from .output import CaptureOutput
from .steno import steno_to_stroke


BLACKBOX_OUTPUT_RX = re.compile("r?['\"]")


def blackbox_setup(blackbox):
    blackbox.output = CaptureOutput()
    blackbox.formatter = Formatter()
    blackbox.formatter.set_output(blackbox.output)
    blackbox.translator = Translator()
    blackbox.translator.set_min_undo_length(100)
    blackbox.translator.add_listener(blackbox.formatter.format)
    blackbox.dictionary = blackbox.translator.get_dictionary()
    blackbox.dictionary.set_dicts([StenoDictionary()])


def blackbox_replay(blackbox, name, test):
    # Hide from traceback on assertions (reduce output size for failed tests).
    __tracebackhide__ = operator.methodcaller("errisinstance", AssertionError)
    definitions, instructions = test.strip().rsplit("\n\n", 1)
    for entry in definitions.split("\n"):
        if entry.startswith(":"):
            _blackbox_replay_action(blackbox, entry[1:])
            continue
        for steno, translation in ast.literal_eval("{" + entry + "}").items():
            blackbox.dictionary.set(normalize_steno(steno), translation)
    # Track line number for a more user-friendly assertion message.
    lines = test.split("\n")
    lnum = len(lines) - 3 - test.rstrip().rsplit("\n\n", 1)[1].count("\n")
    for step in re.split("(?<=[^\\\\])\n", instructions):
        # Mark current instruction's line.
        lnum += 1
        step = step.strip()
        # Support for changing some settings on the fly.
        if step.startswith(":"):
            _blackbox_replay_action(blackbox, step[1:])
            continue
        steno, output = step.split(None, 1)
        steno = list(map(steno_to_stroke, normalize_steno(steno.strip())))
        output = output.strip()
        assert_msg = (
            name
            + "\n"
            + "\n".join(("> " if n == lnum else "  ") + l for n, l in enumerate(lines))
            + "\n"
        )
        if BLACKBOX_OUTPUT_RX.match(output):
            # Replay strokes.
            list(map(blackbox.translator.translate, steno))
            # Check output.
            expected_output = ast.literal_eval(output)
            assert_msg += (
                "   " + repr(blackbox.output.text) + "\n!= " + repr(expected_output)
            )
            assert blackbox.output.text == expected_output, assert_msg
        elif output.startswith("raise "):
            expected_exception = output[6:].strip()
            try:
                list(map(blackbox.translator.translate, steno))
            except Exception as e:
                exception_class = e.__class__.__name__
            else:
                exception_class = "None"
            assert_msg += "   " + exception_class + "\n!= " + expected_exception
            assert exception_class == expected_exception, assert_msg
        else:
            raise ValueError("invalid output:\n%s" % output)


def _blackbox_replay_action(blackbox, action_spec):
    action, *args = shlex.split(action_spec)
    if action == "start_attached":
        assert not args
        blackbox.formatter.start_attached = True
    elif action == "spaces_after":
        assert not args
        blackbox.formatter.set_space_placement("After Output")
    elif action == "spaces_before":
        assert not args
        blackbox.formatter.set_space_placement("Before Output")
    elif action == "system":
        assert len(args) == 1
        system.setup(args[0])
    else:
        raise ValueError("invalid action:\n%r" % action_spec)


def blackbox_test(cls_or_fn):
    # If a class is decorated, we wrap all its test_ methods.
    if inspect.isclass(cls_or_fn):
        # Create a subclass to avoid modifying the original class
        class WrapperClass(cls_or_fn):
            pass

        # Iterate over all attributes in the class.
        for name in dir(WrapperClass):
            # Only wrap methods whose name starts with "test_"
            if name.startswith("test_"):
                fn = getattr(WrapperClass, name)
                # Recursively apply the decorator to each test function.
                new_fn = blackbox_test(fn)
                setattr(WrapperClass, name, new_fn)
        # Return the wrappped class
        return WrapperClass

    else:
        # For a single test function:
        name = cls_or_fn.__name__
        # Dedent the function's docstring which contains the blackbox test script.
        test = textwrap.dedent(cls_or_fn.__doc__)

        @functools.wraps(cls_or_fn)
        def wrapper(bb, *args, **kwargs):
            # Set up the blackbox environment:
            # - This initializes the output capture, formatter, translator,
            #   and dictionary used for replaying strokes.
            blackbox_setup(bb)
            # Execute the original test function.
            cls_or_fn(bb, *args, **kwargs)
            # Replay the test instructions captured in the dedented docstring.
            # This verifies that the outputs match what is expected.
            blackbox_replay(bb, name, test)

        # Return the wrapper function that runs setup, the test, and replay.
        return wrapper



import os
import subprocess
import site
import sys
import itertools

from plover.plugins_manager import global_registry, local_registry
from plover.plugins_manager.utils import running_under_virtualenv


def pip(args, stdin=None, stdout=None, stderr=None, **kwargs):
    cmd = [
        sys.executable,
        "-m",
        "plover.plugins_manager.pip_wrapper",
        "--disable-pip-version-check",
    ]
    env = dict(os.environ)
    # Make sure user plugins are handled
    # even if user site is not enabled.
    if not running_under_virtualenv() and not site.ENABLE_USER_SITE:
        pypath = env.get("PYTHONPATH")
        if pypath is None:
            pypath = []
        else:
            pypath = pypath.split(os.pathsep)
        pypath.insert(0, site.USER_SITE)
        env["PYTHONPATH"] = os.pathsep.join(pypath)
    command = args.pop(0)
    if command == "check":
        cmd.append("check")
    elif command == "install":
        cmd.extend(
            (
                "install",
                "--upgrade-strategy=only-if-needed",
            )
        )
        if not running_under_virtualenv():
            cmd.append("--user")
    elif command == "uninstall":
        cmd.append("uninstall")
    elif command == "list":
        cmd.extend(
            (
                "list",
                "--format=columns",
            )
        )
    else:
        raise ValueError("invalid command: %s" % command)
    cmd.extend(args)
    return subprocess.Popen(
        cmd, env=env, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs
    )


def list_plugins(freeze=False):
    installed_plugins = local_registry.list_plugins()
    if freeze:
        available_plugins = {}
    else:
        available_plugins = global_registry.list_plugins()
    for name, installed, available in sorted(
        (name, installed_plugins.get(name, []), available_plugins.get(name, []))
        for name in set(itertools.chain(installed_plugins, available_plugins))
    ):
        latest = available[-1] if available else None
        current = installed[-1] if installed else None
        info = latest or current
        if freeze:
            if current:
                print("%s==%s" % (current.name, current.version))
            continue
        print("%s (%s)  - %s" % (info.name, info.version, info.summary))
        if current:
            print("  INSTALLED: %s" % current.version)
            if latest:
                print("  LATEST:    %s" % latest.version)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    if args[0] == "list_plugins":
        assert len(args) <= 2
        if len(args) > 1:
            assert args[1] == "--freeze"
            freeze = True
        else:
            freeze = False
        sys.exit(list_plugins(freeze=freeze))
    proc = pip(args)
    sys.exit(proc.wait())


if __name__ == "__main__":
    main()
